"""
Terrible functions I want to get rid off.
"""


def get_dimensions_encoding(n_bins, batchsize):
    """
    Returns a dimensions tuple for 2,3 and 4 dimensional data.
    :param int batchsize: Batchsize that is used in hdf5_batch_generator().
    :param tuple n_bins: Declares the number of bins for each dimension (x,y,z).
                        If a dimension is equal to 1, it means that the dimension should be left out.
    :return: tuple dimensions: 2D, 3D or 4D dimensions tuple (integers).
    """
    n_bins_x, n_bins_y, n_bins_z, n_bins_t = n_bins[0], n_bins[1], n_bins[2], n_bins[3]
    if n_bins_x == 1:
        if n_bins_y == 1:
            print('Using 2D projected data without dimensions x and y')
            dimensions = (batchsize, n_bins_z, n_bins_t, 1)
        elif n_bins_z == 1:
            print('Using 2D projected data without dimensions x and z')
            dimensions = (batchsize, n_bins_y, n_bins_t, 1)
        elif n_bins_t == 1:
            print('Using 2D projected data without dimensions x and t')
            dimensions = (batchsize, n_bins_y, n_bins_z, 1)
        else:
            print('Using 3D projected data without dimension x')
            dimensions = (batchsize, n_bins_y, n_bins_z, n_bins_t, 1)

    elif n_bins_y == 1:
        if n_bins_z == 1:
            print('Using 2D projected data without dimensions y and z')
            dimensions = (batchsize, n_bins_x, n_bins_t, 1)
        elif n_bins_t == 1:
            print('Using 2D projected data without dimensions y and t')
            dimensions = (batchsize, n_bins_x, n_bins_z, 1)
        else:
            print('Using 3D projected data without dimension y')
            dimensions = (batchsize, n_bins_x, n_bins_z, n_bins_t, 1)

    elif n_bins_z == 1:
        if n_bins_t == 1:
            print('Using 2D projected data without dimensions z and t')
            dimensions = (batchsize, n_bins_x, n_bins_y, 1)
        else:
            print('Using 3D projected data without dimension z')
            dimensions = (batchsize, n_bins_x, n_bins_y, n_bins_t, 1)

    elif n_bins_t == 1:
        print('Using 3D projected data without dimension t')
        dimensions = (batchsize, n_bins_x, n_bins_y, n_bins_z, 1)

    else:
        # print 'Using full 4D data'
        dimensions = (batchsize, n_bins_x, n_bins_y, n_bins_z, n_bins_t)

    return dimensions



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import os
import shutil
import h5py
from unittest import TestCase
from keras.models import Model
from keras.layers import Dense, Input, Concatenate

from orcanet.core import Organizer
from orcanet.utilities.nn_utilities import load_zero_center_data, get_layer_output


class TetstZeroCenter(TestCase):
    def setUp(self):
        # Pathes
        # Temporary output folder
        self.temp_dir = os.path.join(os.path.dirname(__file__),
                                     ".temp", "nn_utilities")
        self.output_folder = os.path.join(self.temp_dir, "model")
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")

        config_file = os.path.join(self.data_folder, "config_test.toml")

        # Make sure temp dir does either not exist or is empty
        if os.path.exists(self.temp_dir):
            assert len(os.listdir(self.temp_dir)) == 0
        else:
            os.makedirs(self.temp_dir)

        # Make dummy data
        self.train_inp_A = (
            self.temp_dir + "/train1_A.h5",
            self.temp_dir + "/train2_A.h5"
        )
        self.train_inp_B = (
            self.temp_dir + "/train1_B.h5",
            self.temp_dir + "/train2_B.h5",
        )

        self.train_pathes = {
            "testing_input_A": self.train_inp_A,
            "testing_input_B": self.train_inp_B,
        }
        self.shape = (3, 3, 3)

        make_dummy_data(self.train_inp_A[0], self.train_inp_A[1], self.shape, mode=1)
        make_dummy_data(self.train_inp_B[0], self.train_inp_B[1], self.shape, mode=2)

        orga = Organizer(self.output_folder, config_file=config_file)
        orga.cfg._train_files = self.train_pathes
        orga.cfg._list_file = "test.toml"
        orga.cfg.zero_center_folder = self.temp_dir

        self.orga = orga

    def tearDown(self):
        """ Remove the .temp directory. """
        shutil.rmtree(self.temp_dir)

    def test_load_zero_center(self):
        """ Calculate the zero center image and check if it works properly. """
        orga = self.orga
        xs_mean = load_zero_center_data(orga, logging=False)

        target = {
            "testing_input_A": np.ones(self.shape) * 0.25,
            "testing_input_B": np.ones(self.shape) * 0.75,
        }

        self._check_dict_ndarray(xs_mean, target)

        file_A = orga.cfg.zero_center_folder + "/" + orga.cfg._list_file + \
            '_input_testing_input_A.npz'
        file_B = orga.cfg.zero_center_folder + "/" + orga.cfg._list_file + \
            '_input_testing_input_B.npz'
        used_files_A = np.load(file_A)['zero_center_used_ip_files']
        used_files_B = np.load(file_B)['zero_center_used_ip_files']
        self.assertTrue(np.array_equal(used_files_A, self.train_inp_A))
        self.assertTrue(np.array_equal(used_files_B, self.train_inp_B))

        # test loading of saved xs_mean
        os.remove(file_B)
        xs_mean = load_zero_center_data(orga, logging=False)

        self._check_dict_ndarray(xs_mean, target)
        used_files_A = np.load(file_A)['zero_center_used_ip_files']
        used_files_B = np.load(file_B)['zero_center_used_ip_files']
        self.assertTrue(np.array_equal(used_files_A, self.train_inp_A))
        self.assertTrue(np.array_equal(used_files_B, self.train_inp_B))

    def _check_dict_ndarray(self, val, target):
        self.assertSetEqual(set(val.keys()), set(target.keys()))
        for key in val.keys():
            np.testing.assert_array_almost_equal(val[key], target[key])


class TestModel(TestCase):
    def setUp(self):
        self.layer_names = ["inp1", "inp2", "mid", "outp"]

        inp1 = Input(shape=(2, ), name=self.layer_names[0])
        inp2 = Input(shape=(3, ), name=self.layer_names[1])

        conc = Concatenate()([inp1, inp2])

        intermed = Dense(2,
                         name=self.layer_names[2],
                         bias_initializer="Ones",
                         kernel_initializer="Ones")(conc)
        outp = Dense(1, name=self.layer_names[3])(intermed)

        self.model = Model([inp1, inp2], outp)

    def test_get_layer_output_list_input(self):
        samples = [np.ones((2, 2)), np.ones((2, 3))]

        args = {
            "model": self.model,
            "samples": samples,
            "layer_name": self.layer_names[2],
            "mode": "test",
        }
        value = get_layer_output(**args)
        target = [[6., 6.], [6., 6.]]
        np.testing.assert_array_equal(value, target)

    def test_get_layer_output_dict_input(self):
        samples = {self.layer_names[0]: np.ones((2, 2)),
                   self.layer_names[1]: np.ones((2, 3))}

        args = {
            "model": self.model,
            "samples": samples,
            "layer_name": self.layer_names[2],
            "mode": "test",
        }
        value = get_layer_output(**args)
        target = [[6., 6.], [6., 6.]]
        np.testing.assert_array_equal(value, target)


def make_dummy_data(filepath1, filepath2, shape, mode=1):
    """
    Make two dummy data files.

    mode 1:
      Make a total of 100 ones vs 300 zeroes of dummy data over two files.
      -- > xs_mean = 0.25
    mode 2:
      Make a total of 300 ones vs 100 zeroes of dummy data over two files.
      -- > xs_mean = 0.75

    Parameters
    ----------
    filepath1 : str
        Path to file 1.
    filepath2 : str
        Path to file 2.
    shape : tuple
        Shape of the data, not including sample dimension.

    """
    xs1 = np.ones((150,) + shape)
    xs2 = np.ones((250,) + shape)

    if mode == 1:
        xs1[100:, ...] = 0
        xs2[:, ...] = 0

    elif mode == 2:
        xs2[-100:, ...] = 0

    else:
        raise AssertionError

    dtypes = [('event_id', '<f8'), ('particle_type', '<f8'), ('energy', '<f8'),
              ('is_cc', '<f8'), ('bjorkeny', '<f8'), ('dir_x', '<f8'),
              ('dir_y', '<f8'), ('dir_z', '<f8'), ('time_interaction', '<f8'),
              ('run_id', '<f8'), ('vertex_pos_x', '<f8'), ('vertex_pos_y', '<f8'),
              ('vertex_pos_z', '<f8'), ('time_residual_vertex', '<f8'),
              ('prod_ident', '<f8'), ('group_id', '<i8')]
    ys1 = np.ones((150, 16)).ravel().view(dtype=dtypes)
    ys2 = np.ones((250, 16)).ravel().view(dtype=dtypes)

    for xs, ys, filepath in ((xs1, ys1, filepath1), (xs2, ys2, filepath2)):
        h5f = h5py.File(filepath, 'w')
        h5f.create_dataset('x', data=xs, dtype='uint8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)
        h5f.close()



"""
Multi-gpu code for Keras/TF.
From https://github.com/avolkov1/keras_experiments
"""


import sys
import numpy as np
import tensorflow as tf

from keras.backend import tensorflow_backend as tfb
from keras.backend.tensorflow_backend import (get_session, is_sparse)

import atexit
atexit.register(tfb.clear_session)
# FIXME: The monkey patch of Function results in an error message at the end of
#     the session. Message:
#         Exception AttributeError: "'NoneType' object has no attribute
#         'TF_DeleteStatus'" in <bound method Session.__del__ of
#         <tensorflow.python.client.session.Session object at 0x7fff6c3b3f90>>
#        ignored
#     Everything seems to work fine though.
#     To inhibit the eror run at the end of main script: KB.clear_session()
#     Workaround is to register clear_session at exit.


__all__ = ('patch',)


# Replacement Function to support enqueue ops.
class Function(object):
    """Runs a computation graph.
    # Arguments
        inputs: Feed placeholders to the computation graph.
        outputs: Output tensors to fetch.
        updates: Additional update ops to be run at function call.
        name: a name to help users identify what this function does.
        enqueue_ops: List of ops to be run at funtion call for enqueue'ing.
    """

    def __init__(self, inputs, outputs, updates=None, name=None,
                 enqueue_ops=None, **session_kwargs):
        self._enqueue_ops = enqueue_ops or []
        updates = updates or []
        if not isinstance(inputs, (list, tuple)):
            raise TypeError('`inputs` to a TensorFlow backend function '
                            'should be a list or tuple.')
        if not isinstance(outputs, (list, tuple)):
            raise TypeError('`outputs` of a TensorFlow backend function '
                            'should be a list or tuple.')
        if not isinstance(updates, (list, tuple)):
            raise TypeError('`updates` in a TensorFlow backend function '
                            'should be a list or tuple.')
        self.inputs = list(inputs)
        self.outputs = list(outputs)
        with tf.control_dependencies(self.outputs):
            updates_ops = []
            for update in updates:
                if isinstance(update, tuple):
                    p, new_p = update
                    updates_ops.append(tf.assign(p, new_p))
                else:
                    # assumed already an op
                    updates_ops.append(update)
            self.updates_op = tf.group(*updates_ops)
        self.name = name
        self.session_kwargs = session_kwargs

    def __call__(self, inputs):
        if not isinstance(inputs, (list, tuple)):
            raise TypeError('`inputs` should be a list or tuple.')
        feed_dict = {}
        for tensor, value in zip(self.inputs, inputs):
            if is_sparse(tensor):
                sparse_coo = value.tocoo()
                indices = np.concatenate((np.expand_dims(sparse_coo.row, 1),
                                          np.expand_dims(sparse_coo.col, 1)),
                                         1)
                value = (indices, sparse_coo.data, sparse_coo.shape)
            feed_dict[tensor] = value
        session = get_session()
        enqueue_ops = self._enqueue_ops
        neops = len(enqueue_ops)
        updated = session.run(enqueue_ops + self.outputs + [self.updates_op],
                              feed_dict=feed_dict,
                              **self.session_kwargs)
        nouts = len(self.outputs)

        # return updated[:len(self.outputs)]
        return updated[neops:nouts + neops]


def patch():
    '''Monkey-patch Keras tensorflow_backend.Function'''
    print('\nMONKEY PATCHING KERAS tensorflow_backend.Function\n',
          file=sys.stderr)
tfb.Function = Function


"""
Use orga.predict with a parser.

Usage:
    parser_orcapred.py FOLDER LIST CONFIG MODEL
    parser_orcapred.py (-h | --help)

Arguments:
    FOLDER  Path to the folder where everything gets saved to, e.g. the
            summary.txt, the plots, the trained models, etc.
    LIST    A .toml file which contains the pathes of the training and
            validation files. An example can be found in
            examples/example_list.toml
    CONFIG  A .toml file which sets up the training. An example can be
            found in examples/example_config.toml. The possible parameters
            are listed in core.py in the class Configuration.
    MODEL   Path to a .toml file with infos about a model.
            An example can be found in examples/example_model.toml.

Options:
    -h --help                       Show this screen.

"""
from docopt import docopt
import toml

from orcanet.core import Organizer
from orcanet_contrib.eval_nn import make_performance_plots
from orcanet_contrib.orca_handler_util import update_objects


def orca_pred(output_folder, list_file, config_file, model_file):
    """
    Run orga.predict with predefined ModelBuilder networks using a parser.

    Parameters
    ----------
    output_folder : str
        Path to the folder where everything gets saved to, e.g. the summary
        log file, the plots, the trained models, etc.
    list_file : str
        Path to a list file which contains pathes to all the h5 files that
        should be used for training and validation.
    config_file : str
        Path to a .toml file which overwrites some of the default settings
        for training and validating a model.
    model_file : str
        Path to a file with parameters to build a model of a predefined
        architecture with OrcaNet.

    """
    # Set up the Organizer with the input data
    orga = Organizer(output_folder, list_file, config_file)

    # When predicting with a orga model, the right modifiers and custom
    # objects need to be given
    update_objects(orga, model_file)
    # TODO suboptimal:
    label_modifier = toml.load(model_file)["orca_modifiers"]["label_modifier"]

    # Per default, a prediction will be done for the model with the
    # highest epoch and filenumber.
    pred_filename = orga.predict(epoch=-1, fileno=-1)

    plots_folder = orga.io.get_subfolder(name='plots')
    make_performance_plots(pred_filename, label_modifier, plots_folder)


def parse_input():
    """ Run the orca_pred function with a parser. """
    args = docopt(__doc__)
    output_folder = args['FOLDER']
    list_file = args['LIST']
    config_file = args['CONFIG']
    model_file = args['MODEL']
    orca_pred(output_folder, list_file, config_file, model_file)


if __name__ == '__main__':
    parse_input()



"""
Scripts for writing the logfiles.
"""

import numpy as np
import os
import keras as ks
from datetime import datetime
from shutil import move


class TrainfileLogger:
    def __init__(self, log_file, column_names):
        """
        For writing the training log file in a nice format.

        Parameters
        ----------
        log_file : opened file
            The logfile.
        column_names : List
            A list of column names for the file.

        """
        # Minimum width of the cells in characters.
        self.minimum_cell_width = 11
        # Precision to which floats are rounded if they appear in data.
        self.float_precision = 6

        self.log_file = log_file
        self.column_names = column_names
        self._widths = None

    def level_file(self):
        """
        Make file with only the head lines.

        Existing file will be overwritten.
        """
        headline, widths = self._gen_line_str(self.column_names)
        vline = ["-" * width for width in widths]
        vertical_line = self._gen_line_str(vline, widths, seperator="-+-")[0]
        self.log_file.write(headline + "\n")
        self.log_file.write(vertical_line + "\n")

        self._widths = widths

    def write_line(self, values):
        """
        Write a line with data to the file.

        Parameters
        ----------
        values : List
            The data, in the same order as the column names.

        """
        if self._widths is None:
            raise ValueError("Can not log: .level_file has to be called first")
        if len(values) != len(self.column_names):
            raise ValueError("Can not log: Expected {} values, but got "
                             "{}".format(len(self.column_names), len(values)))

        line = self._gen_line_str(values, self._widths)[0]
        self.log_file.write(line + "\n")

    def _gen_line_str(self, data, widths=None, seperator=" | "):
        line, widths = gen_line_str(data,
                                    widths=widths,
                                    seperator=seperator,
                                    float_precision=self.float_precision,
                                    minimum_cell_width=self.minimum_cell_width)
        return line, widths


def gen_line_str(data, widths=None, seperator=" | ", float_precision=4, minimum_cell_width=9):
    """
    Generate a line in nice human readable format,
    consisting of multiple spaced and seperated cells.

    Parameters
    ----------
    data : tuple
        Strings or floats of what is in each cell. It must be in the
        same order and have the same length as the column names.
    widths : List or None
        Optional: The width of every cell. If None, will set it
        automatically, depending on the data.
        If widths is given, but what is given in data is wider than
        the width, the cell will expand without notice. Must have the
        same length as the column names.
    seperator : str
        String that seperates two adjacent cells.
    float_precision : int
        Precision to which floats are rounded if they appear in data.
        The length of the resulting numbercan be up to 5 characters longer
        than this value (due to . and e-09)

    minimum_cell_width : int
        Minimum width of the cells in characters.

    Returns
    -------
    line : str
        The line.
    new_widths : List
        The widths of the cells.

    """
    if widths is None:
        new_widths = []
    else:
        new_widths = widths

    line = ""
    for i, entry in enumerate(data):
        # no seperator before the first entry
        if i == 0:
            sep = ""
        else:
            sep = seperator

        # If entry is a number, round to given precision and make it a string
        if not isinstance(entry, str):
            entry = format(float(entry), "."+str(float_precision)+"g")

        if widths is None:
            cell_width = max(minimum_cell_width, len(entry))
            new_widths.append(cell_width)
        else:
            cell_width = widths[i]

        cell_cont = format(entry, "<"+str(cell_width))

        line += "{seperator}{entry}".format(seperator=sep, entry=cell_cont,)

    return line, new_widths


class SummaryLogger:
    """
    For writing the summary logfile made during training.
    """
    def __init__(self, orga, model):
        """
        Init with the training files in orga, and metrics in the model.

        Parameters
        ----------
        orga : object Organizer
            Contains all the configurable options in the OrcaNet scripts.
        model : ks.model.Model or None
            Keras model containing the metrics to plot.

        """
        # Minimum width of the cells in characters.
        self.minimum_cell_width = 9
        # Precision to which floats are rounded if they appear in data.
        self.float_precision = 4

        self.logfile_name = orga.cfg.output_folder + 'summary.txt'
        self.temp_filepath = orga.cfg.output_folder + "/.temp_summary.txt"
        self.orga = orga

        self.metric_names = model.metrics_names
        self.column_names = self._get_column_names()

    def write_line(self, epoch_float, lr, history_train=None, history_val=None):
        """
        Write a line to the summary.txt file in every trained model folder.

        Parameters
        ----------
        epoch_float : float
            The current epoch and fileno as a float.
        lr : float
            The current learning rate of the model.
        history_train : dict
            Dict containing the history of the training, averaged over files.
            Keys: Metric names, e.g. "loss", "accuracy", ...
            Values: Value of the metric during validation as a float.
        history_val : dict or None
            Dict of validation losses for all the metrics, averaged over
            all validation files.
            Keys: Metric names, e.g. "loss", "accuracy", ...
            Values: Value of the metric during validation as a float.

        """
        if history_val is None and history_train is None:
            raise ValueError(
                "Can not summary log when both train and val history are None")

        widths = self._init_writing()

        # Format the content: (Epoch, LR, train_1, val_1, ...)
        data = [epoch_float, lr]
        for i, metric_name in enumerate(self.metric_names):
            if history_train is None:
                data.append("nan")
            else:
                data.append(history_train[metric_name])
            if history_val is None:
                data.append("nan")
            else:
                data.append(history_val[metric_name])

        # if the epoch is already in the file, its line will get updated
        update_line = False
        summary_data = self.orga.history.get_summary_data()
        if len(summary_data) > 0:
            last_line = summary_data[-1]
            if last_line["Epoch"] == data[0]:
                # merge arrays but ignore LR
                data = merge_arrays(last_line, data, exclude=1)
                update_line = True

        line = self._gen_line_str(data, widths)[0]
        self._save_line(line, update_line)

    def _get_column_names(self):
        column_names = ["Epoch", "LR", ]
        for i, metric_name in enumerate(self.metric_names):
            column_names.append("train_" + str(metric_name))
            column_names.append("val_" + str(metric_name))
        column_names = tuple(column_names)

        if os.path.isfile(self.logfile_name):
            # if summary exists already, check if model metrics match
            file_column_names = self.orga.history.get_column_names()
            if not set(column_names) == set(file_column_names):
                raise NameError(
                    "Can not log to summary: column names differ (from model: "
                    "{}, from summary file: {}".format(column_names,
                                                       file_column_names))
            column_names = file_column_names

        return column_names

    def _save_line(self, line, update=False):
        """ Write a line in the summary file. If update, overwrite the last line.
        """
        if not update:
            with open(self.logfile_name, 'a+') as logfile:
                logfile.write(line + "\n")

        else:
            # replace last line by rewriting whole summary file :-(
            with open(self.logfile_name, "r") as old_file:
                lines = old_file.readlines()
            lines[-1] = line + "\n"

            # make new summary file as a temp
            with open(self.temp_filepath, 'w') as temp_file:
                for i, old_line in enumerate(lines):
                    temp_file.write(old_line)

            # Remove original file
            os.remove(self.logfile_name)
            # Move new file
            move(self.temp_filepath, self.logfile_name)

    def _init_writing(self):
        """
        Get the widths of the columns, and write the head if the file is new.

        The widths have the length of the metric names, but at least the
        self.minimum_cell_width.

        Returns
        -------
        widths : list
            The width of every cell in characters.

        """
        headline, widths = self._gen_line_str(self.column_names)
        if not os.path.isfile(self.logfile_name) or \
                os.stat(self.logfile_name).st_size == 0:

            vline = ["-" * width for width in widths]
            vertical_line = self._gen_line_str(vline, widths, seperator="-+-")[0]
            with open(self.logfile_name, 'a+') as logfile:
                logfile.write(headline + "\n")
                logfile.write(vertical_line + "\n")

        return widths

    def _gen_line_str(self, data, widths=None, seperator=" | "):
        line, widths = gen_line_str(data,
                                    widths=widths,
                                    seperator=seperator,
                                    float_precision=self.float_precision,
                                    minimum_cell_width=self.minimum_cell_width)
        return line, widths


def merge_arrays(base, supp, exclude=None):
    """
    Fill nans in a list with values from another list.

    Parameters
    ----------
    base : List
    supp : List
    exclude : List or int
        Which indices to ignore.

    Returns
    -------
    np.array

    """
    try:
        iter(exclude)
    except TypeError:
        exclude = [exclude]

    for i in range(len(base)):
        if exclude is not None and i in exclude:
            continue

        if base[i] == supp[i]:
            continue

        elif base[i] == "nan" or np.isnan(base[i]):
            base[i] = supp[i]

        elif supp[i] == "nan" or np.isnan(supp[i]):
            continue

        else:
            raise ValueError(
                "Cannot merge arrays at index {}: Base {}, supplement {}"
                .format(i, base[i], supp[i]))
    return base


class BatchLogger(ks.callbacks.Callback):
    """
    Write logfiles during training.

    Averages the losses of the model over some number of batches,
    and then writes that in a line in the logfile.
    The Batch_float entry in the logfiles gives the absolute position
    of the batch in the epoch (i.e. taking all files into account).
    This class is intended to be used only for one epoch = one file.

    """
    def __init__(self, orga, epoch):
        """

        Parameters
        ----------
        orga : object Organizer
            Contains all the configurable options in the OrcaNet scripts.
        epoch : tuple
            Epoch and file number.

        """
        ks.callbacks.Callback.__init__(self)

        self.epoch_number = epoch[0]
        self.f_number = epoch[1]

        # settings (read from orga)
        self.display = orga.cfg.train_logger_display
        self.flush = orga.cfg.train_logger_flush
        self.logfile_name = '{}/log_epoch_{}_file_{}.txt'.format(
            orga.io.get_subfolder("train_log", create=True),
            self.epoch_number, self.f_number)
        self.batchsize = orga.cfg.batchsize
        self.file_sizes = np.array(orga.io.get_file_sizes("train"))

        # get the total no of batches over all files (not just the current one)
        # This is for calculating the batch_float number in the logs
        file_batches = np.ceil(self.file_sizes / self.batchsize)
        self.total_batches = np.sum(file_batches)
        # no of batches seen in previous files
        if self.f_number == 1:
            self.previous_batches = 0.
        elif self.f_number > 1:
            self.previous_batches = np.cumsum(file_batches)[self.f_number - 2]
        else:
            raise AssertionError("f_number not >= 1 ({})".format(self.f_number))

        self.seen = None
        self.lines = None
        self.cum_metrics = None
        self.file = None
        self._stored_metrics = False
        self._logger = None

    def on_epoch_begin(self, epoch, logs=None):
        # no of seen batches in this epoch
        self.seen = 0
        # list of stored lines, so that multiple can be written at once
        self.lines = []
        # store the various metrices to be able to average over multiple batches
        self.cum_metrics = {}
        for metric in self.model.metrics_names:
            self.cum_metrics[metric] = 0
        self.file = open(self.logfile_name, "w")
        self._write_head()

    def on_batch_end(self, batch, logs=None):
        # self.params:
        #   {'epochs': 1, 'steps': 50, 'verbose': 1, 'do_validation': False,
        #    'metrics': ['loss', 'dx_loss', 'dx_err_loss',
        #    'val_loss', 'val_dx_loss', 'val_dx_err_loss']}
        # logs:
        #   {'batch': 7, 'size': 5, 'loss': 2.06344,
        #    'dx_loss': 0.19809794, 'dx_err_loss': 0.08246058}
        logs = logs or {}

        self.seen += 1
        for metric in self.model.metrics_names:
            self.cum_metrics[metric] += logs.get(metric)
        if not self._stored_metrics:
            self._stored_metrics = True

        if self.seen % self.display == 0:
            self._write_line()

        if self.flush != -1 and self.display % self.flush == 0:
            self._flush_file()

    def on_epoch_end(self, batch, logs=None):
        # on epoch end here means that this is called after one fit_generator
        # loop in Keras is finished, so after one file in our case.
        if self._stored_metrics:
            self._write_line()
        self.file.close()

    def _write_line(self):
        """ Write a line with the metrics for current status and reset metrics.
        """
        # The fraction is shifted by self.display / 2., so that it is in
        # the middle of the samples
        batch_frctn = self.epoch_number - 1 + (
                self.previous_batches + self.seen
                - self.display / 2.) / self.total_batches

        line_data = [self.seen, batch_frctn]
        for metric in self.model.metrics_names:
            line_data.append(self.cum_metrics[metric] / self.display)
            self.cum_metrics[metric] = 0
        self._logger.write_line(line_data)
        self._stored_metrics = False

    def _flush_file(self):
        self.file.flush()
        os.fsync(self.file.fileno())

    def _write_head(self):
        """ write column names for all losses / metrics """
        column_names = ['Batch','Batch_float']
        for i, metric in enumerate(self.model.metrics_names):
            column_names.append(metric)
        self._logger = TrainfileLogger(self.file, column_names)
        self._logger.level_file()


def log_start_training(orga):
    """
    Whenever the orga.train function is run, this logs all the input parameters
    in the full log file.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.

    """
    lines = []
    log = lines.append

    log('-'*60)
    time = datetime.now().strftime('%Y-%m-%d  %H:%M:%S')
    log('-'*19 + " {} ".format(time) + '-'*19)
    log("New training run started with the following configuration:\n")

    log("Output folder:\t" + orga.cfg.output_folder)
    log("List file path:\t" + orga.cfg.get_list_file() + "\n")

    log("Given trainfiles in the .list file:")
    for input_name, input_files in orga.cfg.get_files("train").items():
        log(" " + input_name + ":")
        [log("\t" + input_file) for input_file in input_files]

    log("\nGiven validation files in the .list file:")
    for input_name, input_files in orga.cfg.get_files("val").items():
        log(" " + input_name + ":")
        [log("\t" + input_file) for input_file in input_files]

    log("\nNon-default settings used:")
    for key, value in vars(orga.cfg).items():
        if key == "output_folder" or key.startswith("_") \
                or value == orga.cfg.default_values.get(key):
            continue
        log("   {}:\t{}".format(key, value))

    """
    log("\nDefault settings used:")
    for key, value in vars(orga.cfg).items():
        defaults = orga.cfg.get_defaults()
        if value == defaults.get(key):
            log("   {}:\t{}".format(key, value))
    """

    log("")
    orga.io.print_log(lines)


def log_start_validation(orga):
    """ Log filenames used for validation. """
    line = "Validation"
    orga.io.print_log(line)
    orga.io.print_log("-" * len(line))
    lines = ['Inputs and files:', ]
    for input_name, input_files in orga.io.get_local_files("val").items():
        line = "   " + input_name + ":\t"
        for i, input_file in enumerate(input_files):
            if i != 0:
                line += ", "
            line += os.path.basename(input_file)
        lines.append(line)
    orga.io.print_log(lines)


# class TensorBoardWrapper(ks.callbacks.TensorBoard):
#     """Up to now (05.10.17), Keras doesn't accept TensorBoard callbacks with validation data that is fed by a generator.
#      Supplying the validation data is needed for the histogram_freq > 1 argument in the TB callback.
#      Without a workaround, only scalar values (e.g. loss, accuracy) and the computational graph of the model can be saved.
#
#      This class acts as a Wrapper for the ks.callbacks.TensorBoard class in such a way,
#      that the whole validation data is put into a single array by using the generator.
#      Then, the single array is used in the validation steps. This workaround is experimental!"""
#     def __init__(self, batch_gen, nb_steps, **kwargs):
#         super(TensorBoardWrapper, self).__init__(**kwargs)
#         self.batch_gen = batch_gen # The generator.
#         self.nb_steps = nb_steps   # Number of times to call next() on the generator.
#
#     def on_epoch_end(self, epoch, logs):
#         # Fill in the `validation_data` property.
#         # After it's filled in, the regular on_epoch_end method has access to the validation_data.
#         imgs, tags = None, None
#         for s in range(self.nb_steps):
#             ib, tb = next(self.batch_gen)
#             if imgs is None and tags is None:
#                 imgs = np.zeros(((self.nb_steps * ib.shape[0],) + ib.shape[1:]), dtype=np.float32)
#                 tags = np.zeros(((self.nb_steps * tb.shape[0],) + tb.shape[1:]), dtype=np.uint8)
#             imgs[s * ib.shape[0]:(s + 1) * ib.shape[0]] = ib
#             tags[s * tb.shape[0]:(s + 1) * tb.shape[0]] = tb
#         self.validation_data = [imgs, tags, np.ones(imgs.shape[0]), 0.0]
#         return super(TensorBoardWrapper, self).on_epoch_end(epoch, logs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""""""

import os
import numpy as np
import h5py
from orcanet_contrib.plotting.bg_classifier import select_class
from orcanet_contrib.plotting.utils import select_track_shower


def cut_summary_file():
    """

    """
    col_names = {}
    fpath_col_names = '/home/saturn/capn/mppi033h/Data/standard_reco_files/pid_result_shiftedVertexEventSelection_withDeepLearningTrackScore_column_names.txt'
    meta_filepath = '/home/saturn/capn/mppi033h/Data/standard_reco_files/pid_result_shiftedVertexEventSelection_withDeepLearningTrackScore.meta'
    savepath = '/home/saturn/capn/mppi033h/Data/standard_reco_files/summary_file_cut.h5'

    with open(fpath_col_names) as f_col_names:
        cols_list = f_col_names.read().splitlines()

    for i, col in enumerate(cols_list):
        col_names[col] = i

    selected_cols = ('bjorkeny', 'dir_x', 'dir_y', 'dir_z', 'energy', 'event_id', 'frame_index', 'is_cc', 'is_neutrino', 'n_files_gen',
                     'pos_x', 'pos_y', 'pos_z', 'run_id', 'time', 'type', 'Erange_min', 'dusj_best_DusjOrcaUsingProbabilitiesFinalFit_BjorkenY',
                     'dusj_dir_x', 'dusj_dir_y', 'dusj_dir_z', 'dusj_pos_x', 'dusj_pos_y', 'dusj_pos_z', 'dusj_time',
                     'dusj_is_good', 'dusj_energy_corrected', 'gandalf_dir_x',
                     'gandalf_dir_y', 'gandalf_dir_z', 'gandalf_pos_x', 'gandalf_pos_y', 'gandalf_pos_z', 'gandalf_time',
                     'gandalf_is_good', 'gandalf_energy_corrected', 'dusj_is_selected', 'gandalf_is_selected',
                     'gandalf_loose_is_selected', 'muon_score', 'track_score', 'noise_score')

    usecols = [col_names[col_name] for col_name in selected_cols]
    #TODO fix doesnt work, wrong col naming and assignment

    print('Loading the summary file content')
    dtype = {}
    dtype['names'] = selected_cols
    dtype['formats'] = (np.float64, ) * len(selected_cols)
    summary_file_arr = np.loadtxt(meta_filepath, delimiter=' ', usecols=usecols.sort(), dtype=dtype)
    # summary_file_arr = np.loadtxt(meta_filepath, delimiter=' ', usecols=usecols.sort(), names=cols_list)  # a lot slower and uses tons more memory

    f = h5py.File(savepath, 'w')
    print('Saving the summary file content')
    f.create_dataset('summary_array', data=summary_file_arr, compression='gzip', compression_opts=1)
    f.close()


def make_pred_file():
    """

    """
    f_in = h5py.File('/home/saturn/capn/mppi033h/Data/standard_reco_files/summary_file_cut.h5', 'r')

    # bg classifier selections
    print('Making files for bg classifier')
    s = f_in['summary_array']
    dusj_is_selected = s['dusj_is_selected']
    gandalf_loose_is_selected = s['gandalf_loose_is_selected']
    selection_classifier_bg = np.logical_or(dusj_is_selected, gandalf_loose_is_selected)

    mc_info_sel_bg = get_mc_info(s, selection=selection_classifier_bg)
    save_surviving_evt_info_to_npy(mc_info_sel_bg, '/home/saturn/capn/mppi033h/Data/event_selections/evt_selection_bg_classifier.npy')

    pred_bg = get_pred(s, 'bg_classifier_2_class', selection=selection_classifier_bg)
    save_dsets_to_h5_file(mc_info_sel_bg, pred_bg,
                          '/home/saturn/capn/mppi033h/Data/standard_reco_files/pred_file_bg_classifier_2_class.h5')

    # ts classifier
    print('Making files for ts classifier')
    print(f_in['summary_array']['is_neutrino'])
    is_neutrino = f_in['summary_array']['is_neutrino'] == 1

    s = f_in['summary_array'][is_neutrino]
    dusj_is_selected = s['dusj_is_selected']
    gandalf_loose_is_selected = s['gandalf_loose_is_selected']
    selection_classifier_ts = np.logical_or(dusj_is_selected, gandalf_loose_is_selected)

    mc_info_sel_ts = get_mc_info(s, selection=selection_classifier_ts)
    save_surviving_evt_info_to_npy(mc_info_sel_ts, '/home/saturn/capn/mppi033h/Data/event_selections/evt_selection_ts_classifier.npy')

    pred_ts = get_pred(s, 'ts_classifier', selection=selection_classifier_ts)
    save_dsets_to_h5_file(mc_info_sel_ts, pred_ts,
                          '/home/saturn/capn/mppi033h/Data/standard_reco_files/pred_file_ts_classifier.h5')

    # regression
    print('Making files for regression')
    is_neutrino = f_in['summary_array']['is_neutrino'] == 1
    s = f_in['summary_array'][is_neutrino]
    dusj_is_selected = s['dusj_is_selected']
    gandalf_is_selected = s['gandalf_is_selected']
    selection_regr = np.logical_or(dusj_is_selected, gandalf_is_selected)

    mc_info_sel_regr = get_mc_info(s, selection=selection_regr)
    save_surviving_evt_info_to_npy(mc_info_sel_ts,
                                   '/home/saturn/capn/mppi033h/Data/event_selections/evt_selection_regression.npy')

    pred_regr = get_pred(s, 'regression_e_dir_vtx_by', selection=selection_regr)
    save_dsets_to_h5_file(mc_info_sel_regr, pred_regr,
                          '/home/saturn/capn/mppi033h/Data/standard_reco_files/pred_file_regression.h5')

    f_in.close()

    # TODO make print how many events have been thrown away


def get_mc_info(s, selection=None):
    """

    Parameters
    ----------
    s
    selection

    Returns
    -------
    mc_info

    """
    if selection is not None:
        s = s[selection]

    cols_s_out = [('event_id', 'event_id'), ('type', 'particle_type'), ('energy', 'energy'),
                  ('is_cc', 'is_cc'), ('bjorkeny', 'bjorkeny'), ('dir_x', 'dir_x'),
                  ('dir_y', 'dir_y'), ('dir_z', 'dir_z'), ('time', 'time_interaction'),
                  ('run_id', 'run_id'), ('pos_x', 'vertex_pos_x'), ('pos_y', 'vertex_pos_y'),
                  ('pos_z', 'vertex_pos_z')]

    dtypes = [(tpl[1], s[tpl[0]].dtype) for tpl in cols_s_out]

    # make prod_ident col
    prod_ident = np.zeros(s['run_id'].shape[0], dtype=np.float64)

    is_mupage = select_class('mupage', s['type'], s['is_cc'])
    is_random_noise = select_class('random_noise', s['type'], s['is_cc'])
    is_neutrino = np.invert(np.logical_or(is_mupage, is_random_noise))
    is_neutrino_low_e = np.logical_and(s['Erange_min'], is_neutrino)
    is_neutrino_high_e = np.logical_and(np.invert(is_neutrino_low_e), is_neutrino)

    prod_ident[is_mupage] = 3
    prod_ident[is_random_noise] = 4
    prod_ident[is_neutrino_low_e] = 2
    prod_ident[is_neutrino_high_e] = 1
    assert np.count_nonzero(prod_ident == 0) == 0

    cols_s_out.append((None, 'prod_id'))
    dtypes.append(('prod_id', np.float64))

    n_evts = s['energy'].shape[0]
    mc_info = np.empty(n_evts, dtype=dtypes)
    for tpl in cols_s_out:
        if tpl[1] == 'prod_id':
            mc_info['prod_id'] = prod_ident
        else:
            mc_info[tpl[1]] = s[tpl[0]]

    return mc_info


def save_surviving_evt_info_to_npy(mc_info, savepath):
    """

    Parameters
    ----------
    mc_info
    savepath

    """
    ax = np.newaxis
    arr = np.concatenate([mc_info['run_id'][:, ax], mc_info['event_id'][:, ax],
                          mc_info['prod_id'][:, ax], mc_info['particle_type'][:, ax],
                          mc_info['is_cc'][:, ax]], axis=1)

    np.save(savepath, arr)


def get_pred(s, mode, selection=None):
    """

    Parameters
    ----------
    s
    mode
    selection

    Returns
    -------
    pred

    """
    if selection is not None:
        s = s[selection]

    if mode == 'bg_classifier':
        raise ValueError('bg_classifier not yet implemented')
        # need 3 values, prob mupage, prob random_noise, prob neutrino

    elif mode == 'bg_classifier_2_class':
        # need 2 values, prob neutrino, prob not neutrino

        prob_not_neutrino = s['noise_score']
        prob_neutrino = 1 - prob_not_neutrino

        dtypes_pred = [('prob_not_neutrino', np.float64), ('prob_neutrino', np.float64)]
        n_evts = prob_not_neutrino.shape[0]
        pred = np.empty(n_evts, dtype=dtypes_pred)

        pred['prob_neutrino'] = prob_neutrino
        pred['prob_not_neutrino'] = prob_not_neutrino

    elif mode == 'ts_classifier':
        # need 2 values, prob shower, prob track

        prob_track = s['track_score']
        prob_shower = 1 - prob_track

        dtypes_pred = [('prob_track', np.float64), ('prob_shower', np.float64)]
        n_evts = prob_track.shape[0]
        pred = np.empty(n_evts, dtype=dtypes_pred)

        pred['prob_track'] = prob_track
        pred['prob_shower'] = prob_shower

    elif mode == 'regression_e_dir_vtx_by':
        is_track, is_shower = select_track_shower(s['type'], s['is_cc'])
        s_shower, s_track = s[is_shower], s[is_track]

        # energy
        energy_shower, energy_track = s_shower['dusj_energy_corrected'], s_track['gandalf_energy_corrected']
        energy = np.concatenate([energy_shower, energy_track], axis=0)

        # bjorkeny
        bjorkeny_shower = s_shower['dusj_best_DusjOrcaUsingProbabilitiesFinalFit_BjorkenY']
        bjorkeny_track = np.full(s_track.shape, 0.360366)  # not available for gandalf, use median of by_true of training dataset
        bjorkeny = np.concatenate([bjorkeny_shower, bjorkeny_track], axis=0)

        # dir and vtx
        labels_shower = ['dusj_dir_x', 'dusj_dir_y', 'dusj_dir_z', 'dusj_pos_x', 'dusj_pos_y', 'dusj_pos_z', 'dusj_time']
        labels_track = ['gandalf_dir_x', 'gandalf_dir_y', 'gandalf_dir_z', 'gandalf_pos_x', 'gandalf_pos_y', 'gandalf_pos_z', 'gandalf_time']
        dir_vtx_shower, dir_vtx_track = {}, {}

        for i, label in enumerate(labels_shower):
            dir_vtx_shower[i] = s_shower[label]

        for i, label in enumerate(labels_track):
            dir_vtx_track[i] = s_track[label]

        dir_vtx = {}
        for i in range(len(dir_vtx_shower)):
            dir_vtx[i] = np.concatenate([dir_vtx_shower[i], dir_vtx_track[i]], axis=0)

        labels_pred_dset = ['pred_energy', 'pred_dir_x', 'pred_dir_y', 'pred_dir_z', 'pred_bjorkeny', 'pred_vtx_x',
                            'pred_vtx_y', 'pred_vtx_z', 'pred_vtx_t']

        dtypes_pred = [(label_pred_name, np.float64) for label_pred_name in labels_pred_dset]
        n_evts = energy.shape[0]
        pred = np.empty(n_evts, dtype=dtypes_pred)

        pred['pred_energy'] = energy
        pred['pred_bjorkeny'] = bjorkeny
        pred['pred_dir_x'], pred['pred_dir_y'], pred['pred_dir_z'] = dir_vtx[0], dir_vtx[1], dir_vtx[2]
        pred['pred_vtx_x'], pred['pred_vtx_y'], pred['pred_vtx_z'], pred['pred_vtx_t'] = dir_vtx[3], dir_vtx[4], dir_vtx[5], dir_vtx[6]

    return pred


def save_dsets_to_h5_file(mc_info, pred, savepath):
    """

    Parameters
    ----------
    mc_info
    pred
    savepath

    """
    f = h5py.File(savepath, 'w')
    f.create_dataset('mc_info', data=mc_info, compression='gzip', compression_opts=1)
    f.create_dataset('pred', data=pred, compression='gzip', compression_opts=1)
    f.close()


if __name__ == '__main__':
    if os.path.exists('/home/saturn/capn/mppi033h/Data/standard_reco_files/summary_file_cut.h5') is False:
        cut_summary_file()  # do on tinyfat or woody with 32g, qsub -I -l nodes=1:ppn=4:sl32g,walltime=05:00:00

    make_pred_file()









from keras.models import Model, clone_model
from keras.layers import Input, Dense, Concatenate, Dropout
import numpy as np
from keras import backend as K


def build_double_inp(compile=False):
    inp_1 = Input((1,), name="inp_0")
    inp_2 = Input((1,), name="inp_1")

    x = Concatenate()([inp_1, inp_2])
    x = Dense(3)(x)

    output_1 = Dense(1, name="out_0")(x)
    output_2 = Dense(2, name="out_1")(x)

    test_model = Model((inp_1, inp_2), (output_1, output_2))

    if compile:
        test_model.compile("adam", loss={"out_0": "mse", "out_1": "mse"},
                           metrics={"out_0": "mae", "out_1": "mae"})

    return test_model


def build_single_inp():
    inp = Input((2,), name="inp")

    x = Dense(3)(inp)

    output_1 = Dense(1, name="out_0")(x)
    output_2 = Dense(2, name="out_1")(x)

    test_model = Model(inp, (output_1, output_2))
    return test_model


def get_xs(model, batchsize=1):
    """ Get dummy data fitting a model. """
    shapes = model.input_shape
    if len(model.input_names) == 1:
        shapes = [shapes, ]
    xs = {model.input_names[i]: np.ones([batchsize, ] + list(shapes[i][1:]))
          for i in range(len(model.input_names))}
    return xs


def dropout_test():
    def dropout_model(rate=0.):
        inp = Input((10,))
        out = Dropout(rate)(inp)
        model = Model(inp, out)
        return model

    def get_layer_output(model, xs, which=-1):
        l_out = K.function([model.layers[0].input, K.learning_phase()],
                           [model.layers[which].output])
        # output in train mode = 1
        layer_output = l_out([xs, 1])[0]
        return layer_output

    model0 = dropout_model(0.)
    model1 = dropout_model(0.99)
    xs = np.ones((3, 10))

    print("no drop\n", get_layer_output(model0, xs))
    print("\nmax drop\n", get_layer_output(model1, xs))
    model1.layers[-1].rate = 0.
    print("\nchanged max drop to zero\n", model1.layers[-1].get_config())
    print(get_layer_output(model1, xs))
    model1_clone = clone_model(model1)
    print("\n clone changed model\n", get_layer_output(model1_clone, xs))


def get_structured_array():
    x = np.array([(9, 81.0), (3, 18),  (3, 18),  (3, 18)],
                 dtype=[('inp_0', 'f4'), ('inp_1', 'f4')])
    return x


def get_dict():
    x = {"inp_0": np.array([9, 3]), "inp_1": np.array([81, 18])}
    return x


def transf_arr(x):
    xd = {name: x[name] for name in x.dtype.names}
    return xd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
TODO
"""
import numpy as np
import toml

from orcanet_contrib.custom_objects import get_custom_objects


def update_objects(orga, model_file):
    """
    Update the organizer for using the model.

    Look up and load in the respective sample-, label-, and dataset-
    modifiers, as well as the custom objects.
    Will assert that the respective objects have not already been set
    to a non-default value (nothing is overwritten).

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model_file : str
        Path to a toml file which has the infos about which modifiers
        to use.

    """
    file_content = toml.load(model_file)
    orca_modifiers = file_content["orca_modifiers"]

    sample_modifier = orca_modifiers.get("sample_modifier")
    label_modifier = orca_modifiers.get("label_modifier")
    dataset_modifier = orca_modifiers.get("dataset_modifier")

    if sample_modifier is not None:
        assert orga.cfg.sample_modifier is None, \
            "Can not set sample modifier: Has already been set: " \
            "{}".format(orga.cfg.sample_modifier)
    if label_modifier is not None:
        assert orga.cfg.label_modifier is None, \
            "Can not set label modifier: Has already been set: " \
            "{}".format(orga.cfg.label_modifier)
    if dataset_modifier is not None:
        assert orga.cfg.dataset_modifier is None, \
            "Can not set dataset modifier: Has already been set: " \
            "{}".format(orga.cfg.dataset_modifier)
    assert orga.cfg.custom_objects is None, \
        "Can not set custom objects: Have already been set: " \
        "{}".format(orga.cfg.custom_objects)

    if sample_modifier is not None:
        print("Using orga sample modifier: ", sample_modifier)
        orga.cfg.sample_modifier = orca_sample_modifiers(sample_modifier)
    if label_modifier is not None:
        print("Using orga label modifier: ", label_modifier)
        orga.cfg.label_modifier = orca_label_modifiers(label_modifier)
    if dataset_modifier is not None:
        print("Using orga dataset modifier: ", dataset_modifier)
        orga.cfg.dataset_modifier = orca_dataset_modifiers(dataset_modifier)
    print("Using orga custom objects")
    orga.cfg.custom_objects = get_custom_objects()


def orca_sample_modifiers(name):
    """
    Returns one of the sample modifiers used for Orca networks.

    They will permute columns, and/or add permuted columns to xs.

    The input to the functions is:
        xs_files : dict
            Dict that contains the input samples from the file(s).
            The keys are the names of the inputs in the toml list file.
            The values are a single batch of data from each corresponding file.

    The output is:
        xs_layer : dict
            Dict that contains the input samples for a Keras NN.
            The keys are the names of the input layers of the network.
            The values are a single batch of data for each input layer.

    Parameters
    ----------
    name : None/str
        Name of the sample modifier to return.

    Returns
    -------
    sample_modifier : function
        The sample modifier function.

    """
    # assuming input is bxyzt
    xyzt_permute = {'yzt-x': (0, 2, 3, 4, 1),
                    'xyt-z': (0, 1, 2, 4, 3),
                    't-xyz': (0, 4, 1, 2, 3),
                    'tyz-x': (0, 4, 2, 3, 1)}

    if name in xyzt_permute:
        def swap_columns(xs_files):
            # Transpose dimensions
            xs_layer = dict()
            keys = list(xs_files.keys())
            xs_layer[keys[0]] = np.transpose(xs_files, xyzt_permute[name])
            return xs_layer
        sample_modifier = swap_columns

    elif name == 'xyz-t_and_yzt-x':
        def sample_modifier(xs_files):
            # Use xyz-t, and also transpose it to yzt-x and use that, too.
            xs_layer = dict()
            xs_layer['xyz-t'] = xs_files['xyz-t']
            xs_layer['yzt-x'] = np.transpose(xs_files['xyz-t'], xyzt_permute['yzt-x'])
            return xs_layer

    elif name == 'xyz-t_and_yzt-x_multi_input_single_train_tight-1_tight-2':
        def sample_modifier(xs_files):
            # Use xyz-t in two different time cuts, and also transpose them to yzt-x and use these, too.
            xs_layer = dict()
            xs_layer['xyz-t_tight-1'] = xs_files['xyz-t_tight-1']
            xs_layer['xyz-t_tight-2'] = xs_files['xyz-t_tight-2']
            xs_layer['yzt-x_tight-1'] = np.transpose(xs_files['xyz-t_tight-1'],
                                                     xyzt_permute['yzt-x'])
            xs_layer['yzt-x_tight-2'] = np.transpose(xs_files['xyz-t_tight-2'],
                                                     xyzt_permute['yzt-x'])
            return xs_layer

    elif name == 'xyz-t_and_xyz-c_single_input':
        def sample_modifier(xs_files):
            # Concatenate xyz-t and xyz-c to a single input
            xs_layer = dict()
            xs_layer['xyz-t_and_xyz-c_single_input'] = np.concatenate(
                [xs_files['xyz-t'], xs_files['xyz-c']], axis=-1)
            return xs_layer

    else:
        raise ValueError('Unknown input_type: ' + str(name))

    return sample_modifier


def orca_label_modifiers(name):
    """
    Returns one of the label modifiers used for Orca networks.

    CAREFUL: y_values is a structured numpy array! if you use advanced
    numpy indexing, this may lead to errors. Let's suppose you want to
    assign a particular value to one or multiple elements of the
    y_values array.

    E.g.
    y_values[1]['bjorkeny'] = 5
    This works, since it is basic indexing.

    Likewise,
    y_values[1:3]['bjorkeny'] = 5
    works as well, because basic indexing gives you a view (!).

    Advanced indexing though, gives you a copy.
    So this
    y_values[[1,2,4]]['bjorkeny'] = 5
    will NOT work! Same with boolean indexing, like

    bool_idx = np.array([True,False,False,True,False]) # if len(y_values) = 5
    y_values[bool_idx]['bjorkeny'] = 10
    This will NOT work as well!!

    Instead, use
    np.place(y_values['bjorkeny'], bool_idx, 10)
    This works.

    Parameters
    ----------
    name : str
        Name of the label modifier that should be used.

    Returns
    -------
    label_modifier : function
        The label modifier function.

    """

    if name == 'energy_dir_bjorken-y_vtx_errors':
        def label_modifier(y_values):
            ys = dict()
            particle_type, is_cc = y_values['particle_type'], y_values['is_cc']
            elec_nc_bool_idx = np.logical_and(np.abs(particle_type) == 12,
                                              is_cc == 0)

            # correct energy to visible energy
            visible_energy = y_values[elec_nc_bool_idx]['energy'] * \
                y_values[elec_nc_bool_idx]['bjorkeny']
            # fix energy to visible energy
            np.place(y_values['energy'], elec_nc_bool_idx, visible_energy)
            # set bjorkeny label of nc events to 1
            np.place(y_values['bjorkeny'], elec_nc_bool_idx, 1)

            ys['dx'], ys['dx_err'] = y_values['dir_x'], y_values['dir_x']
            ys['dy'], ys['dy_err'] = y_values['dir_y'], y_values['dir_y']
            ys['dz'], ys['dz_err'] = y_values['dir_z'], y_values['dir_z']
            ys['e'], ys['e_err'] = y_values['energy'], y_values['energy']
            ys['by'], ys['by_err'] = y_values['bjorkeny'], y_values['bjorkeny']

            ys['vx'], ys['vx_err'] = y_values['vertex_pos_x'], y_values['vertex_pos_x']
            ys['vy'], ys['vy_err'] = y_values['vertex_pos_y'], y_values['vertex_pos_y']
            ys['vz'], ys['vz_err'] = y_values['vertex_pos_z'], y_values['vertex_pos_z']
            ys['vt'], ys['vt_err'] = y_values['time_residual_vertex'], y_values['time_residual_vertex']

            for key_label in ys:
                ys[key_label] = ys[key_label].astype(np.float32)
            return ys

    elif name == 'ts_classifier':
        def label_modifier(y_values):
            # for every sample, [0,1] for shower, or [1,0] for track

            # {(12, 0): 0, (12, 1): 1, (14, 1): 2, (16, 1): 3}
            # 0: elec_NC, 1: elec_CC, 2: muon_CC, 3: tau_CC
            # label is always shower, except if muon-CC
            ys = dict()
            particle_type, is_cc = y_values['particle_type'], y_values['is_cc']
            is_muon_cc = np.logical_and(np.abs(particle_type) == 14, is_cc == 1)
            is_not_muon_cc = np.invert(is_muon_cc)

            batchsize = y_values.shape[0]
            # categorical [shower, track] -> [1,0] = shower, [0,1] = track
            categorical_ts = np.zeros((batchsize, 2), dtype='bool')

            categorical_ts[:, 0] = is_not_muon_cc
            categorical_ts[:, 1] = is_muon_cc

            ys['ts_output'] = categorical_ts.astype(np.float32)
            return ys

    elif name == 'bg_classifier':
        def label_modifier(y_values):
            # for every sample, [1,0,0] for neutrinos, [0,1,0] for mupage
            # and [0,0,1] for random_noise
            # particle types: mupage: np.abs(13), random_noise = 0, neutrinos =
            ys = dict()
            particle_type = y_values['particle_type']
            is_mupage = np.abs(particle_type) == 13
            is_random_noise = np.abs(particle_type == 0)
            is_not_mupage_nor_rn = np.invert(np.logical_or(is_mupage,
                                                           is_random_noise))

            batchsize = y_values.shape[0]
            categorical_bg = np.zeros((batchsize, 3), dtype='bool')

            categorical_bg[:, 0] = is_not_mupage_nor_rn
            categorical_bg[:, 1] = is_mupage
            categorical_bg[:, 2] = is_random_noise

            ys['bg_output'] = categorical_bg.astype(np.float32)
            return ys

    elif name == 'bg_classifier_2_class':
        def label_modifier(y_values):
            # for every sample, [1,0,0] for neutrinos, [0,1,0] for mupage
            # and [0,0,1] for random_noise
            # particle types: mupage: np.abs(13), random_noise = 0, neutrinos =
            ys = dict()
            particle_type = y_values['particle_type']
            is_mupage = np.abs(particle_type) == 13
            is_random_noise = np.abs(particle_type == 0)
            is_not_mupage_nor_rn = np.invert(np.logical_or(is_mupage,
                                                           is_random_noise))

            batchsize = y_values.shape[0]
            categorical_bg = np.zeros((batchsize, 2), dtype='bool')

            # neutrino
            categorical_bg[:, 0] = is_not_mupage_nor_rn
            # is not neutrino
            categorical_bg[:, 1] = np.invert(is_not_mupage_nor_rn)

            ys['bg_output'] = categorical_bg.astype(np.float32)
            return ys

    else:
        raise ValueError("Unknown output_type: " + str(name))

    return label_modifier


def orca_dataset_modifiers(name):
    """
    Returns one of the dataset modifiers used for predicting with OrcaNet.

    Parameters
    ----------
    name : str
        Name of the dataset modifier that should be used.

    """
    if name == 'bg_classifier':
        def dataset_modifier(mc_info, y_true, y_pred):

            # y_pred and y_true are dicts with keys for each output
            # we only have 1 output in case of the bg classifier
            y_pred = y_pred['bg_output']
            y_true = y_true['bg_output']

            datasets = dict()
            datasets['mc_info'] = mc_info  # is already a structured array

            # make pred dataset
            dtypes = np.dtype([('prob_neutrino', y_pred.dtype),
                               ('prob_muon', y_pred.dtype),
                               ('prob_random_noise', y_pred.dtype)])
            pred = np.empty(y_pred.shape[0], dtype=dtypes)
            pred['prob_neutrino'] = y_pred[:, 0]
            pred['prob_muon'] = y_pred[:, 1]
            pred['prob_random_noise'] = y_pred[:, 2]

            datasets['pred'] = pred

            # make true dataset
            dtypes = np.dtype([('cat_neutrino', y_true.dtype),
                               ('cat_muon', y_true.dtype),
                               ('cat_random_noise', y_true.dtype)])
            true = np.empty(y_true.shape[0], dtype=dtypes)
            true['cat_neutrino'] = y_true[:, 0]
            true['cat_muon'] = y_true[:, 1]
            true['cat_random_noise'] = y_true[:, 2]

            datasets['true'] = true

            return datasets

    elif name == 'bg_classifier_2_class':
        def dataset_modifier(mc_info, y_true, y_pred):

            # y_pred and y_true are dicts with keys for each output
            # we only have 1 output in case of the bg classifier
            y_pred = y_pred['bg_output']
            y_true = y_true['bg_output']

            datasets = dict()  # y_pred is a list of arrays
            datasets['mc_info'] = mc_info  # is already a structured array

            # make pred dataset
            dtypes = np.dtype([('prob_neutrino', y_pred.dtype),
                               ('prob_not_neutrino', y_pred.dtype)])
            pred = np.empty(y_pred.shape[0], dtype=dtypes)
            pred['prob_neutrino'] = y_pred[:, 0]
            pred['prob_not_neutrino'] = y_pred[:, 1]

            datasets['pred'] = pred

            # make true dataset
            dtypes = np.dtype([('cat_neutrino', y_true.dtype),
                               ('cat_not_neutrino', y_true.dtype)])
            true = np.empty(y_true.shape[0], dtype=dtypes)
            true['cat_neutrino'] = y_true[:, 0]
            true['cat_not_neutrino'] = y_true[:, 1]

            datasets['true'] = true

            return datasets

    elif name == 'ts_classifier':
        def dataset_modifier(mc_info, y_true, y_pred):

            # y_pred and y_true are dicts with keys for each output
            # we only have 1 output in case of the ts classifier
            y_pred = y_pred['ts_output']
            y_true = y_true['ts_output']

            datasets = dict()
            datasets['mc_info'] = mc_info  # is already a structured array

            # make pred dataset
            dtypes = np.dtype([('prob_shower', y_pred.dtype),
                               ('prob_track', y_pred.dtype)])
            pred = np.empty(y_pred.shape[0], dtype=dtypes)
            pred['prob_shower'] = y_pred[:, 0]
            pred['prob_track'] = y_pred[:, 1]

            datasets['pred'] = pred

            # make true dataset
            dtypes = np.dtype([('cat_shower', y_true.dtype),
                               ('cat_track', y_true.dtype)])
            true = np.empty(y_true.shape[0], dtype=dtypes)
            true['cat_shower'] = y_true[:, 0]
            true['cat_track'] = y_true[:, 1]

            datasets['true'] = true

            return datasets

    elif name == 'regression_energy_dir_bjorken-y_vtx_errors':
        def dataset_modifier(mc_info, y_true, y_pred):

            datasets = dict()
            datasets['mc_info'] = mc_info  # is already a structured array

            # make pred dataset
            """y_pred and y_true are dicts with keys for each output,
               here, we have 1 key for each regression variable"""

            pred_labels_and_nn_output_names = [('pred_energy', 'e'), ('pred_dir_x', 'dx'), ('pred_dir_y', 'dy'),
                                               ('pred_dir_z', 'dz'), ('pred_bjorkeny', 'by'), ('pred_vtx_x', 'vx'),
                                               ('pred_vtx_y', 'vy'), ('pred_vtx_z', 'vz'), ('pred_vtx_t', 'vt'),
                                               ('pred_err_energy', 'e_err'), ('pred_err_dir_x', 'dx_err'),
                                               ('pred_err_dir_y', 'dy_err'), ('pred_err_dir_z', 'dz_err'),
                                               ('pred_err_bjorkeny', 'by_err'), ('pred_err_vtx_x', 'vx_err'),
                                               ('pred_err_vtx_y', 'vy_err'), ('pred_err_vtx_z', 'vz_err'),
                                               ('pred_err_vtx_t', 'vt_err')]

            dtypes_pred = [(tpl[0], y_pred[tpl[1]].dtype) for tpl in pred_labels_and_nn_output_names]
            n_evts = y_pred['e'].shape[0]
            pred = np.empty(n_evts, dtype=dtypes_pred)

            for tpl in pred_labels_and_nn_output_names:
                pred[tpl[0]] = y_pred[tpl[1]]

            datasets['pred'] = pred

            # make true dataset
            true_labels_and_nn_output_names = [('true_energy', 'e'), ('true_dir_x', 'dx'), ('true_dir_y', 'dy'),
                                               ('true_dir_z', 'dz'), ('true_bjorkeny', 'by'), ('true_vtx_x', 'vx'),
                                               ('true_vtx_y', 'vy'), ('true_vtx_z', 'vz'), ('true_vtx_t', 'vt'),
                                               ('true_err_energy', 'e_err'), ('true_err_dir_x', 'dx_err'),
                                               ('true_err_dir_y', 'dy_err'), ('true_err_dir_z', 'dz_err'),
                                               ('true_err_bjorkeny', 'by_err'), ('true_err_vtx_x', 'vx_err'),
                                               ('true_err_vtx_y', 'vy_err'), ('true_err_vtx_z', 'vz_err'),
                                               ('true_err_vtx_t', 'vt_err')]

            dtypes_true = [(tpl[0], y_true[tpl[1]].dtype) for tpl in true_labels_and_nn_output_names]
            true = np.empty(n_evts, dtype=dtypes_true)

            for tpl in true_labels_and_nn_output_names:
                true[tpl[0]] = y_true[tpl[1]]

            datasets['true'] = true

    else:
        raise ValueError('Unknown dataset modifier: ' + str(name))

    return dataset_modifier


def orca_learning_rates(name, total_file_no):
    """
    Returns one of the learning rate schedules used for Orca networks.

    Parameters
    ----------
    name : str
        Name of the schedule.
    total_file_no : int
        How many files there are to train on.

    Returns
    -------
    learning_rate : function
        The learning rate schedule.

    """
    if name == "triple_decay":
        def learning_rate(n_epoch, n_file):
            """
            Function that calculates the current learning rate based on
            the number of already trained epochs.

            Learning rate schedule: lr_decay = 7% for lr > 0.0003
                                    lr_decay = 4% for 0.0003 >= lr > 0.0001
                                    lr_decay = 2% for 0.0001 >= lr

            Parameters
            ----------
            n_epoch : int
                The number of the current epoch which is used to calculate
                the new learning rate.
            n_file : int
                The number of the current filenumber which is used to
                calculate the new learning rate.

            Returns
            -------
            lr_temp : float
                Calculated learning rate for this epoch.

            """
            n_lr_decays = (n_epoch - 1) * total_file_no + (n_file - 1)
            lr_temp = 0.005  # * n_gpu TODO think about multi gpu lr

            for i in range(n_lr_decays):
                if lr_temp > 0.0003:
                    lr_decay = 0.07  # standard for regression: 0.07, standard for PID: 0.02
                elif 0.0003 >= lr_temp > 0.0001:
                    lr_decay = 0.04  # standard for regression: 0.04, standard for PID: 0.01
                else:
                    lr_decay = 0.02  # standard for regression: 0.02, standard for PID: 0.005
                lr_temp = lr_temp * (1 - float(lr_decay))

            return lr_temp
    else:
        raise NameError("Unknown orca learning rate name", name)

    return learning_rate



from keras.models import Model
from keras.layers import Dense, Flatten, Concatenate, Lambda
from keras import backend as K
from keras.regularizers import l2
import inspect

from orcanet.builder_util.layer_blocks import (
    input_block, ConvBlock, DenseBlock)


class BlockBuilder:
    """
    Can build single-input block-wise sequential neural network.

    Attributes
    ----------
    body_defaults : dict or None
        Default values for all blocks in the body section of the model.
    head_defaults : dict or None
        Default values for all blocks in the head section of the model.

    """

    def __init__(self, body_defaults=None, head_defaults=None):
        """ Set dicts with default values for the layers of the model.
        """
        self._all_blocks = {
            "conv_block": ConvBlock,
            "dense_block": DenseBlock,
        }

        if body_defaults is not None and "kernel_reg" in body_defaults:
            if body_defaults["kernel_reg"] == "l2":
                body_defaults["kernel_reg"] = l2
            else:
                raise NameError(
                    "Unknown kernel_reg: " + str(body_defaults["kernel_reg"]))

        if head_defaults is not None and "kernel_reg" in head_defaults:
            if head_defaults["kernel_reg"] == "l2":
                head_defaults["kernel_reg"] = l2
            else:
                raise NameError(
                    "Unknown kernel_reg: " + str(head_defaults["kernel_reg"]))

        self._check_arguments(body_defaults)
        self._check_arguments(head_defaults)

        self.body_defaults = body_defaults
        self.head_defaults = head_defaults

    def build(self, input_shape, body_configs, head_arch, head_arch_args=None):
        """
        Build the whole model, using the default values when arguments
        are missing in the layer_configs.

        Parameters
        ----------
        input_shape : dict
            Name and shape of the input layer.
        body_configs : list
            List of configurations for the blocks in the body of the model.
            Each element in the list is a dict and will result in one block
            connected to the previous one. The dict has to contain the type
            of the block, as well as any arguments required by that
            specific block type.
        head_arch : str or None
            Specifies the architecture of the head.
        head_arch_args : dict or None
            Required arguments for the given head_arch.

        Returns
        -------
        model : keras model

        """
        if not isinstance(input_shape, dict):
            raise TypeError(
                "input_shapes must be a dict, not ", type(input_shape))
        if not len(input_shape) == 1:
            raise TypeError(
                "input_shapes must have length 1, not ", len(input_shape))

        input_layer = input_block(input_shape)[0]

        x = input_layer
        for layer_config in body_configs:
            x = self.attach_block(x, layer_config)
        conv_output_flat = Flatten()(x)

        if head_arch is None:
            outputs = conv_output_flat
        else:
            outputs = self.attach_output_layers(conv_output_flat,
                                                head_arch, head_arch_args)

        model = Model(inputs=input_layer, outputs=outputs)
        return model

    def attach_block(self, layer, layer_config, is_output=False):
        """
        Attach a block to the given layer based on the layer config.

        Will use the default values given during initialization if they are not
        present in the layer config.

        Parameters
        ----------
        layer : keras layer
            Layer to attach the block to.
        layer_config : dict
            Configuration of the block to attach. The dict has to contain
            the type of the block, as well as any arguments required by that
            specific block.
        is_output : bool
            Whether to use the default values for the head layers or
            the body layers.

        Returns
        -------
        x : keras layer

        """
        if is_output:
            defaults = self.head_defaults
        else:
            defaults = self.body_defaults
        filled = self._with_defaults(layer_config, defaults)
        block = self._get_blocks(filled.pop("type"))
        x = block(**filled)(layer)
        return x

    def attach_output_layers(self, layer, head_arch, head_arch_args):
        """
        Append the head dense layers to the network.

        Parameters
        ----------
        layer : keras layer
            The Keras layer instance to attach the head to.
        head_arch : str
            Specifies the architecture of the head layers.
        head_arch_args : dict
            Required arguments for the given output_arch.

        Returns
        -------
        outputs : list
            List of the keras output layers of the network.

        """
        if head_arch == "categorical":
            # categorical problem
            assert head_arch_args is not None, "No output_kwargs given"
            outputs = self.attach_output_cat(layer, **head_arch_args)

        elif head_arch == "regression_error":
            # regression with error estimation, two outputs for each label
            assert head_arch_args is not None, "No output_kwargs given"
            outputs = self.attach_output_reg_err(layer, **head_arch_args)

        else:
            raise ValueError("Unknown head_arch: " + str(head_arch))

        return outputs

    def attach_output_cat(self, layer, categories, output_name):
        """ Small dense network for multiple categories. """
        x = self.attach_block(
            layer, {"type": "dense_block", "units": 128},
            is_output=True)

        x = self.attach_block(
            x, {"type": "dense_block", "units": 32, "dropout": None},
            is_output=True)

        out = Dense(units=categories, activation='softmax',
                    kernel_initializer='he_normal', name=output_name)(x)
        return [out, ]

    def attach_output_reg_err(self, layer, output_names):
        """ Double network for regression + error estimation. """
        outputs = []

        # Network for the labels
        x = self.attach_block(
            layer, {"type": "dense_block", "units": 128},
            is_output=True)

        x = self.attach_block(
            x, {"type": "dense_block", "units": 32, "dropout": None},
            is_output=True)

        for name in output_names:
            output_label = Dense(units=1, name=name)(x)
            outputs.append(output_label)

        # Network for the errors of the labels
        x_err = Lambda(lambda a: K.stop_gradient(a))(layer)

        x_err = self.attach_block(
            x_err, {"type": "dense_block", "units": 128},
            is_output=True)

        x_err = self.attach_block(
            x_err, {"type": "dense_block", "units": 64},
            is_output=True)

        x_err = self.attach_block(
            x_err, {"type": "dense_block", "units": 32, "dropout": None},
            is_output=True)

        for i, name in enumerate(output_names):
            output_label_error = Dense(units=1, activation='linear',
                                       name=name + '_err_temp')(x_err)
            # Predicted label gets concatenated with its error (needed for loss function)
            output_label_merged = Concatenate(name=name + '_err')([outputs[i],
                                                                   output_label_error])
            outputs.append(output_label_merged)

        return outputs

    def _with_defaults(self, config, defaults):
        """ Make a copy of a layer config and complete it with default values
        for its block, if they are missing in the layer config.
        """
        conf = dict(config)

        if config is not None and "type" in config:
            block_name = config["type"]
        elif defaults is not None and "type" in defaults:
            block_name = defaults["type"]
            conf["type"] = defaults["type"]
        else:
            raise KeyError("No layer block type specified")

        block = self._get_blocks(block_name)
        args = inspect.getfullargspec(block).args
        if "self" in args:
            args.remove("self")

        if defaults is not None:
            for key, val in defaults.items():
                if key in args and key not in conf:
                    conf[key] = val

        return conf

    def _get_blocks(self, name=None):
        """ Get the block class/function depending on the name. """
        if name is None:
            return self._all_blocks
        elif name in self._all_blocks:
            return self._all_blocks[name]
        else:
            raise NameError("Unknown block type: " + str(name))

    def _check_arguments(self, defaults):
        """ Check if given defaults appear in at least one block. """
        if defaults is None:
            return
        # possible arguments for all blocks
        psb_args = ["type", ]
        for block in self._get_blocks().values():
            args = inspect.getfullargspec(block).args
            for arg in args:
                if arg not in psb_args and arg != "self":
                    psb_args.append(arg)

        for t_def in defaults.keys():
            if t_def not in psb_args:
                raise NameError("Unknown argument: " + str(t_def))




'''
'''
import sys

try:
    from io import StringIO
except ImportError:
    # Python 3 compat.
    from io import StringIO


__all__ = ('Capturing',)


class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Core scripts for the OrcaNet package.
"""

import os
import toml
import warnings
import numpy as np
from keras.models import load_model
from keras.utils import plot_model
import keras.backend as kb

from orcanet.backend import make_model_prediction, save_actv_wghts_plot, get_learning_rate, train_model, validate_model
from orcanet.utilities.visualization import update_summary_plot
from orcanet.in_out import IOHandler, HistoryHandler
from orcanet.utilities.nn_utilities import load_zero_center_data, get_auto_label_modifier
from orcanet.logging import log_start_training, SummaryLogger, log_start_validation


class Organizer:
    """
    Core class for working with networks in OrcaNet.

    Attributes
    ----------
    cfg : Configuration
        Contains all configurable options.
    io : orcanet.in_out.IOHandler
        Utility functions for accessing the info in cfg.
    history : orcanet.in_out.HistoryHandler
        For reading and plotting data from the log files created
        during training.

    """

    def __init__(self, output_folder, list_file=None, config_file=None):
        """
        Set the attributes of the Configuration object.

        Instead of using a config_file, the attributes of orga.cfg can
        also be changed directly, e.g. by calling orga.cfg.batchsize.

        Parameters
        ----------
        output_folder : str
            Name of the folder of this model in which everything will be saved,
            e.g., the summary.txt log file is located in here.
            Will be used to load saved files or to save new ones.
        list_file : str or None
            Path to a toml list file with pathes to all the h5 files that should
            be used for training and validation.
            Will be used to extract of samples or labels.
        config_file : str or None
            Path to a toml config file with settings that are used instead of
            the default ones.

        """
        self.cfg = Configuration(output_folder, list_file, config_file)
        self.io = IOHandler(self.cfg)
        self.history = HistoryHandler(self.cfg.output_folder + "summary.txt",
                                      self.io.get_subfolder("train_log"))

        self.xs_mean = None
        self._auto_label_modifier = None
        self._stored_model = None

    def train_and_validate(self, model=None, epochs=None):
        """
        Train a model and validate according to schedule.

        The various settings of this process can be controlled with the
        attributes of orca.cfg.
        The model will be trained on the given data, saved and validated.
        Logfiles of the training are saved in the output folder.
        Plots showing the training and validation history, as well as
        the weights and activations of the network are generated in
        the plots subfolder after every validation.
        The training can be resumed by executing this function again.

        Parameters
        ----------
        model : ks.models.Model or None
            Compiled keras model to use for training and validation. Required
            for the first epoch (the start of training).
            Afterwards, if model is None, the most recent model will be
            loaded automatically to continue the training.
        epochs : int or None
            How many epochs should be trained by running this function.
            None for infinite.

        Returns
        -------
        model : ks.models.Model
            The trained keras model.

        """
        latest_epoch = self.io.get_latest_epoch()

        model = self._get_model(model, logging=True)
        self._stored_model = model

        # check if the validation is missing for the latest fileno
        if latest_epoch is not None:
            state = self.history.get_state()[-1]
            if state["is_validated"] is False and self._val_is_due(latest_epoch):
                    self.validate()

        next_epoch = self.io.get_next_epoch(latest_epoch)
        n_train_files = self.io.get_no_of_files("train")

        trained_epochs = 0
        while epochs is None or trained_epochs < epochs:
            # Train on remaining files
            for file_no in range(next_epoch[1], n_train_files + 1):
                curr_epoch = (next_epoch[0], file_no)
                self.train(model)
                if self._val_is_due(curr_epoch):
                    self.validate()

            next_epoch = (next_epoch[0] + 1, 1)
            trained_epochs += 1

        self._stored_model = None
        return model

    def train(self, model=None):
        """
        Trains a model on the next file.

        The progress of the training is also logged and plotted.

        Parameters
        ----------
        model : ks.models.Model or None
            Compiled keras model to use for training. Required for the first
            epoch (the start of training).
            Afterwards, if model is None, the most recent model will be
            loaded automatically to continue the training.

        Returns
        -------
        history : dict
            The history of the training on this file. A record of training
            loss values and metrics values.

        """
        # Create folder structure
        self.io.get_subfolder(create=True)
        latest_epoch = self.io.get_latest_epoch()

        model = self._get_model(model, logging=True)

        self._set_up(model, logging=True)

        # epoch about to be trained
        next_epoch = self.io.get_next_epoch(latest_epoch)
        next_epoch_float = self.io.get_epoch_float(*next_epoch)

        if latest_epoch is None:
            self.io.check_connections(model)
            log_start_training(self)

        model_path = self.io.get_model_path(*next_epoch)
        if os.path.isfile(model_path):
            raise FileExistsError(
                "Can not train model in epoch {} file {}, this model has "
                "already been saved!".format(*next_epoch))

        smry_logger = SummaryLogger(self, model)

        n_train_files = self.io.get_no_of_files("train")
        lr = get_learning_rate(next_epoch, self.cfg.learning_rate,
                               n_train_files)
        kb.set_value(model.optimizer.lr, lr)

        files_dict = self.io.get_file("train", next_epoch[1])

        line = "Training in epoch {} on file {}/{}".format(
            next_epoch[0], next_epoch[1], n_train_files)
        self.io.print_log(line)
        self.io.print_log("-" * len(line))
        self.io.print_log("Learning rate is at {}".format(
            kb.get_value(model.optimizer.lr)))
        self.io.print_log('Inputs and files:')
        for input_name, input_file in files_dict.items():
            self.io.print_log("   {}: \t{}".format(input_name,
                                                   os.path.basename(
                                                       input_file)))

        history = train_model(self, model, next_epoch, batch_logger=True)

        model.save(model_path)
        smry_logger.write_line(next_epoch_float, lr, history_train=history)

        self.io.print_log('Training results:')
        for metric_name, loss in history.items():
            self.io.print_log("   {}: \t{}".format(metric_name, loss))
        self.io.print_log("Saved model to " + model_path + "\n")

        update_summary_plot(self)

        return history

    def validate(self):
        """
        Validate the most recent saved model on all validation files.

        Will also log the progress, as well as update the summary plot and
        plot weights and activations of the model.

        Returns
        -------
        history : dict
            The history of the validation on all files. A record of validation
            loss values and metrics values.

        """
        latest_epoch = self.io.get_latest_epoch()
        if latest_epoch is None:
            raise ValueError("Can not validate: No saved model found")
        if self._stored_model is None:
            model = self.load_saved_model(*latest_epoch)
        else:
            model = self._stored_model

        self._set_up(model, logging=True)

        epoch_float = self.io.get_epoch_float(*latest_epoch)
        smry_logger = SummaryLogger(self, model)

        log_start_validation(self)

        history = validate_model(self, model)

        self.io.print_log('Validation results:')
        for metric_name, loss in history.items():
            self.io.print_log("   {}: \t{}".format(metric_name, loss))
        self.io.print_log("")
        smry_logger.write_line(epoch_float, np.nan, history_val=history)

        update_summary_plot(self)
        save_actv_wghts_plot(self, model, latest_epoch, samples=self.cfg.batchsize)

        return history

    def predict(self, epoch=-1, fileno=-1):
        """
        Make a prediction if it does not exist yet, and return its filepath.

        Load a model, let it predict on all samples of the validation set
        in the toml list, and save this prediction together with all the
        mc_info as a h5 file in the predictions subfolder.

        Parameters
        ----------
        epoch : int
            The epoch of the model to load for prediction
            Can also give -1 to automatically load the most recent epoch
            found in the main folder.
        fileno : int
            When using multiple files, define the file number for the
            prediction, e.g. 1 for load the model trained on the first file.
            Can also give -1 to automatically load the most recent fileno
            from the given epoch found in the main folder.

        Returns
        -------
        pred_filename : str
            The path to the created prediction file.

        """
        if fileno == -1:
            epoch, fileno = self.io.get_latest_epoch(epoch)
            print("Automatically set epoch to epoch {} file {}.".format(epoch,
                                                                        fileno))

        pred_filename = self.io.get_pred_path(epoch, fileno)

        if os.path.isfile(pred_filename):
            print("Prediction has already been done.")
        else:
            model = self.load_saved_model(epoch, fileno, logging=False)
            self._set_up(model)

            make_model_prediction(self, model, pred_filename, samples=None)

        return pred_filename

    def get_xs_mean(self, logging=False):
        """
        Set and return the zero center image for each list input.

        Requires the cfg.zero_center_folder to be set. If no existing
        image for the given input files is found in the folder, it will
        be calculated and saved by averaging over all samples in the
        train dataset.

        Parameters
        ----------
        logging : bool
            If true, the execution of this function will be logged into the
            full summary in the output folder if called for the first time.

        Returns
        -------
        dict
            Dict of numpy arrays that contains the mean_image of the x dataset
            (1 array per list input).
            Example format:
            { "input_A" : ndarray, "input_B" : ndarray }

        """
        if self.xs_mean is None:
            if self.cfg.zero_center_folder is None:
                raise ValueError("Can not calculate zero center: "
                                 "No zero center folder given")
            self.xs_mean = load_zero_center_data(self, logging=logging)
        return self.xs_mean

    def load_saved_model(self, epoch, fileno, logging=False):
        """
        Load a saved model.

        Parameters
        ----------
        epoch : int
            Epoch of the saved model.
        fileno : int
            Fileno of the saved model.
        logging : bool
            If True, will log this function call into the log.txt file.

        Returns
        -------
        model : keras model

        """
        path_of_model = self.io.get_model_path(epoch, fileno)
        self.io.print_log("Loading saved model: " + path_of_model, logging=logging)
        model = load_model(path_of_model, custom_objects=self.cfg.custom_objects)
        return model

    def _get_model(self, model, logging=False):
        """ Load most recent saved model or use user model. """
        latest_epoch = self.io.get_latest_epoch()

        if latest_epoch is None:
            if model is None:
                raise ValueError("You need to provide a compiled keras model "
                                 "for the start of the training! (You gave None)")
            try:
                plot_model(model,
                           self.io.get_subfolder("plots") + "/model_plot.png")
            except OSError as e:
                warnings.warn("Can not plot model: " + str(e))

        else:
            if model is None:
                model = self.load_saved_model(*latest_epoch, logging=logging)

        return model

    def _set_up(self, model, logging=False):
        """ Necessary setup for training, validating and predicting. """
        if self.cfg.filter_out_tf_garbage:
            os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

        if self.cfg.get_list_file() is None:
            raise ValueError("No files specified. You need to load a toml "
                             "list file with your files before training")

        if self.cfg.label_modifier is None:
            self._auto_label_modifier = get_auto_label_modifier(model)

        if self.cfg.use_scratch_ssd:
            self.io.use_local_node()

        if self.cfg.zero_center_folder is not None:
            self.get_xs_mean(logging)

    def _val_is_due(self, epoch):
        """ True if validation is due on given epoch according to schedule. """
        n_train_files = self.io.get_no_of_files("train")
        val_sched = (epoch[1] == n_train_files) or \
                    (self.cfg.validate_interval is not None and
                     epoch[1] % self.cfg.validate_interval == 0)
        return val_sched


class Configuration(object):
    """
    Contains all the configurable options in the OrcaNet scripts.

    All of these public attributes (the ones without a
    leading underscore) can be changed either directly or with a
    .toml config file via the method update_config().

    Attributes
    ----------
    batchsize : int
        Batchsize that should be used for the training and validation of
        the network.
    callback_train : keras callback or list or None
        Callback or list of callbacks to use during training.
    custom_objects : dict or None
        Optional dictionary mapping names (strings) to custom classes or
        functions to be considered by keras during deserialization of models.
    dataset_modifier : function or None
        For orga.pred: Function that determines which datasets get created
        in the resulting h5 file. If none, every output layer will get one
        dataset each for both the label and the prediction, and one dataset
        containing the mc_info from the validation files.
    filter_out_tf_garbage : bool
        If true, surpresses the tensorflow info logs which usually spam
        the terminal.
    key_samples : str
        The name of the datagroup in the h5 input files which contains
        the samples for the network.
    key_labels : str
        The name of the datagroup in the h5 input files which contains
        the info for the labels.
    label_modifier : function or None
        Operation to be performed on batches of labels read from the input files
        before they are fed into the model. If None is given, all labels with
        the same name as the output layers will be passed to the model as a dict,
        with the keys being the dtype names.
    learning_rate : float or tuple or function
        The learning rate for the training.
        If it is a float, the learning rate will be constantly this value.
        If it is a tuple of two floats, the first float gives the learning rate
        in epoch 1 file 1, and the second float gives the decrease of the
        learning rate per file (e.g. 0.1 for 10% decrease per file).
        You can also give a function, which takes as an input the epoch and the
        file number (in this order), and returns the learning rate.
    max_queue_size : int
        max_queue_size option of the keras training and evaluation generator
        methods. How many batches get preloaded
        from the generator.
    n_events : None or int
        For testing purposes. If not the whole .h5 file should be used for
        training, define the number of samples.
    sample_modifier : function or None
        Operation to be performed on batches of samples read from the input
        files before they are fed into the model.
    shuffle_train : bool
        If true, the order in which batches are read out from the files during
        training are randomized each time they are read out.
    train_logger_display : int
        How many batches should be averaged for one line in the training log files.
    train_logger_flush : int
        After how many lines the training log file should be flushed (updated on
        the disk). -1 for flush at the end of the file only.
    output_folder : str
        Name of the folder of this model in which everything will be saved,
        e.g., the summary.txt log file is located in here.
    use_scratch_ssd : bool
        Only working at HPC Erlangen: Declares if the input files should be
        copied to the node-local SSD scratch space.
    validate_interval : int or None
        Validate the model after this many training files have been trained on
        in an epoch. There will always be a validation at the end of an epoch.
        None for only validate at the end of an epoch.
        Example: validate_interval=3 --> Validate after file 3, 6, 9, ...
    verbose_train : int
        verbose option of keras.model.fit_generator.
        0 = silent, 1 = progress bar, 2 = one line per epoch.
    verbose_val : int
        verbose option of evaluate_generator.
        0 = silent, 1 = progress bar.
    zero_center_folder : None or str
        Path to a folder in which zero centering images are stored.
        If this path is set, zero centering images for the given dataset will
        either be calculated and saved automatically at the start of the
        training, or loaded if they have been saved before.

    """
    # TODO add a clober script that properly deletes models + logfiles
    def __init__(self, output_folder, list_file, config_file):
        """
        Set the attributes of the Configuration object.

        Values are loaded from the given files, if provided. Otherwise, default
        values are used.

        Parameters
        ----------
        output_folder : str
            Name of the folder of this model in which everything will be saved,
            e.g., the summary.txt log file is located in here.
        list_file : str or None
            Path to a toml list file with pathes to all the h5 files that should
            be used for training and validation.
        config_file : str or None
            Path to a toml config file with attributes that are used instead of
            the default ones.

        """
        self.batchsize = 64
        self.learning_rate = 0.001

        self.zero_center_folder = None
        self.validate_interval = None

        self.sample_modifier = None
        self.dataset_modifier = None
        self.label_modifier = None

        self.key_samples = "x"
        self.key_labels = "y"
        self.custom_objects = None
        self.shuffle_train = False

        self.callback_train = None
        self.use_scratch_ssd = False
        self.verbose_train = 1
        self.verbose_val = 0

        self.n_events = None
        self.filter_out_tf_garbage = True
        self.max_queue_size = 10
        self.train_logger_display = 100
        self.train_logger_flush = -1

        self._default_values = dict(self.__dict__)

        # Main folder:
        if output_folder[-1] == "/":
            self.output_folder = output_folder
        else:
            self.output_folder = output_folder+"/"

        # Private attributes:
        self._train_files = None
        self._val_files = None
        self._list_file = None

        # Load the optionally given list and config files.
        if list_file is not None:
            self.import_list_file(list_file)
        if config_file is not None:
            self.update_config(config_file)

    def import_list_file(self, list_file):
        """
        Import the filepaths of the training and validation files from a toml
        list file.

        Parameters
        ----------
        list_file : str
            Path to the toml list file.

        """
        if self._list_file is not None:
            raise ValueError("Can not load list file: Has already been loaded! "
                             "({})".format(self._list_file))

        file_content = toml.load(list_file)
        train_files, validation_files = {}, {}

        # no of train/val files in each input set
        n_train, n_val = [], []
        for input_key, input_values in file_content.items():
            if not len(input_values.keys()) == 2:
                raise ValueError("Wrong input format in toml list file (input {}:"
                                 " {})".format(input_key, input_values))
            if "train_files" not in input_values.keys():
                raise NameError("No train files specified in toml list file")
            if "validation_files" not in input_values.keys():
                raise NameError("No validation files specified in toml list file")

            train_files[input_key] = tuple(input_values["train_files"])
            validation_files[input_key] = tuple(input_values["validation_files"])
            n_train.append(len(train_files[input_key]))
            n_val.append(len(validation_files[input_key]))

        if not n_train.count(n_train[0]) == len(n_train):
            raise ValueError("The specified training inputs do not "
                             "all have the same number of files!")
        if not n_val.count(n_val[0]) == len(n_val):
            raise ValueError("The specified validation inputs do not "
                             "all have the same number of files!")

        self._train_files = train_files
        self._val_files = validation_files
        self._list_file = list_file

    def update_config(self, config_file):
        """
        Update the default cfg parameters with values from a toml config file.

        Parameters
        ----------
        config_file : str
            Path to a toml config file.

        """
        user_values = toml.load(config_file)["config"]
        for key in user_values:
            if hasattr(self, key):
                setattr(self, key, user_values[key])
            else:
                raise AttributeError(
                    "Unknown attribute {} in config file ".format(key))

    def get_list_file(self):
        """
        Returns the path to the list file that was used to set the training
        and validation files. None if no list file has been used.

        """
        return self._list_file

    def get_files(self, which):
        """
        Get the training or validation file paths for each list input set.

        Returns
        -------
        dict
            A dict containing the paths to the training or validation files on
            which the model will be trained on. Example for the format for
            two input sets with two files each:
                    {
                     "input_A" : ('path/to/set_A_file_1.h5', 'path/to/set_A_file_2.h5'),
                     "input_B" : ('path/to/set_B_file_1.h5', 'path/to/set_B_file_2.h5'),
                    }

        """
        if which == "train":
            if self._train_files is None:
                raise AttributeError("No train files have been specified!")
            return self._train_files
        elif which == "val":
            if self._val_files is None:
                raise AttributeError("No validation files have been specified!")
            return self._val_files
        else:
            raise NameError("Unknown fileset name ", which)

    @property
    def default_values(self):
        """ The default values for all settings. """
        return self._default_values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for making plots for track-shower classifiers.
"""

import numpy as np
from matplotlib import pyplot as plt
from orcanet_contrib.plotting.utils import get_event_selection_mask, select_ic


# -- Utility functions -- #

def get_latex_code_for_ptype_str(class_name):
    """
    Returns the mpl latex code for a certain particle type class (e.g. muon-CC).

    Parameters
    ----------
    class_name : str
        String that specifies the particle type class.

    Returns
    -------
    label : str
        class_name, reformated into a latex readable string.

    """
    labels = {'muon-CC': r'$\nu_{\mu}-CC$', 'a_muon-CC': r'$\overline{\nu}_{\mu}-CC$', 'elec-CC': r'$\nu_{e}-CC$',
              'a_elec-CC': r'$\overline{\nu}_{e}-CC$', 'elec-NC': r'$\nu_{e}-NC$',
              'a_elec-NC': r'$\overline{\nu}_{e}-NC$', 'tau-CC': r'$\nu_{\tau}-CC$',
              'a_tau-CC': r'$\overline{\nu}_{\tau}-CC$'}
    label = labels[class_name]

    return label

# -- Utility functions -- #


# -- Functions for making energy to accuracy plots -- #

def make_e_to_acc_plot_ts(pred_file, title, savefolder, plot_range=(1, 100),
                          cuts=None, prob_threshold_shower=0.5):
    """
    Makes a track-shower step plot for energy to "Fraction of events classified as track" for different neutrino types.

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the track-shower classification predictions of a nn model.
    title : str
        Title that should be used for the plot.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    plot_range : tuple(int,int)
        Tuple that specifies the X-Range of the plot.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.
    prob_threshold_shower : float
        Sets the lower threshold for when an event is classified as a shower based on the nn shower probability.

    """
    y_pred, y_true, mc_info = pred_file['pred'], pred_file['true'], pred_file['mc_info']
    nn_pred_correct = get_nn_pred_correct_info(y_pred, y_true, prob_threshold_shower=prob_threshold_shower)
    print_accuracy(nn_pred_correct, print_text='Accuracy of the T/S classifier without any event selection: ')

    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        nn_pred_correct = nn_pred_correct[evt_sel_mask]
        print_accuracy(nn_pred_correct, print_text='Accuracy of the T/S classifier with the event selection "'
                                                   + cuts + '": ')

    fig, axes = plt.subplots()

    make_step_plot_e_acc_class('muon-CC', mc_info, nn_pred_correct, axes, plot_range=plot_range, linestyle='-', color='b')
    make_step_plot_e_acc_class('a_muon-CC', mc_info, nn_pred_correct, axes, plot_range=plot_range, linestyle='--', color='b')
    make_step_plot_e_acc_class('elec-CC', mc_info, nn_pred_correct, axes, invert=True, plot_range=plot_range, linestyle='-', color='r')
    make_step_plot_e_acc_class('a_elec-CC', mc_info, nn_pred_correct, axes, invert=True, plot_range=plot_range, linestyle='--', color='r')
    make_step_plot_e_acc_class('elec-NC', mc_info, nn_pred_correct, axes, invert=True, plot_range=plot_range, linestyle='-', color='saddlebrown')
    make_step_plot_e_acc_class('a_elec-NC', mc_info, nn_pred_correct, axes, invert=True, plot_range=plot_range, linestyle='--', color='saddlebrown')
    make_step_plot_e_acc_class('tau-CC', mc_info, nn_pred_correct, axes, invert=True, plot_range=plot_range, linestyle='-', color='g')
    make_step_plot_e_acc_class('a_tau-CC', mc_info, nn_pred_correct, axes, invert=True, plot_range=plot_range, linestyle='--', color='g')

    axes.legend(loc='center right', ncol=2)

    x_ticks_major = np.arange(0, 101, 10)
    y_ticks_major = np.arange(0, 1.1, 0.1)
    plt.xticks(x_ticks_major)
    plt.minorticks_on()

    plt.xlabel('Energy [GeV]')
    plt.ylabel('Fraction of events classified as track')
    plt.ylim((0, 1.05))
    plt.yticks(y_ticks_major)
    title = plt.title(title)
    title.set_position([.5, 1.04])
    plt.grid(True, zorder=0, linestyle='dotted')

    plt.text(0.05, 0.92, 'KM3NeT Preliminary', transform=axes.transAxes, weight='bold')

    plt.savefig(savefolder + '/ts_e_to_acc_3-100GeV.pdf')
    plt.savefig(savefolder + '/ts_e_to_acc_3-100GeV.png', dpi=600)

    x_ticks_major = np.arange(0, 101, 5)
    plt.xticks(x_ticks_major)
    plt.xlim((0, 40))
    plt.savefig(savefolder + '/ts_e_to_acc_3-40GeV.pdf')
    plt.savefig(savefolder + '/ts_e_to_acc_3-40GeV.png', dpi=600)

    plt.close()


def get_nn_pred_correct_info(y_pred, y_true, prob_threshold_shower=0.5):
    """
    Function that checks if the predictions by the neural network are correct or not for the binary T/S classification.

    Parameters
    ----------
    y_pred : ndarray(ndim=2)
        Structured array with the nn probabilities for the shower and the track class.
    y_true : ndarray(ndim=2)
        Structured array with the one hot true values for the shower and the track class.
    prob_threshold_shower : float
        Lower threshold, which sets the minimum value, at which an event is considered a shower based
        on the predicted nn shower probability.

    Returns
    -------
    nn_pred_correct : ndarray(ndim=1)
        Boolean array which specifies, if the predictions of the nn for each event are correct or not.

    """
    if len(y_pred.dtype) > 2:
        raise ValueError('The check if a T/S prediction of a nn is correct is only available for '
                         'binary categorization problems and not for problems with more than two classes!')

    is_shower_pred = (y_pred['prob_shower'] > prob_threshold_shower)
    is_shower_true = (y_true['is_shower'] > 0)

    nn_pred_correct = np.logical_and(is_shower_pred, is_shower_true)

    return nn_pred_correct


def print_accuracy(nn_pred_correct, print_text='Accuracy of the T/S classifier: '):
    """
    Prints the T/S accuracy of a nn based on the nn_pred_correct info.

    Parameters
    ----------
    nn_pred_correct : ndarray(ndim=1)
        Boolean array which specifies, if the predictions of the nn for each event are correct or not.

    print_text : str
        String that should be used in printing before printing the results.

    """
    n_correct = np.count_nonzero(nn_pred_correct)
    n_total = nn_pred_correct.shape[0]
    accuracy = n_correct / float(n_total)

    print(print_text, '\n', str(accuracy * 100), ', based on ', n_total, ' events')


def select_class(class_name, ptype, is_cc):
    """
    Returns a boolean array which specifies, which rows in the ptype & is_cc 1d arrays belong to a single
    class, specified by the class_name.

    Parameters
    ----------
    class_name : str
        String that specifies the class that should be selected for the boolean flag output.
    ptype : ndarray(ndim=1)
        Array with particle_types of some events.
    is_cc : ndarray(ndim=1)
        Array with is_cc of some events.

    Returns
    -------
    is_class : ndarray(ndim=1)
        Boolean flag, which specifies if each row belongs to the class specified by the class_name.

    """
    particle_types = {'muon-CC': (14, 1), 'a_muon-CC': (-14, 1), 'elec-CC': (12, 1),
                      'a_elec-CC': (-12, 1), 'elec-NC': (12, 0), 'a_elec-NC': (-12, 0),
                      'tau-CC': (16, 1), 'a_tau-CC': (-16, 1)}

    try:
        is_class = np.logical_and(ptype == particle_types[class_name][0], is_cc == particle_types[class_name][1])
    except KeyError:
        raise ValueError('The class ' + str(class_name) + ' is not available in the particle_types dictionary.')

    return is_class


def make_step_plot_e_acc_class(class_name, mc_info, nn_pred_correct, axes, invert=False, plot_range=(3, 100),
                               linestyle='-', color='b'):
    """
    Plots 1D "energy" to "Accuracy" step plots to an axes object for a certain particle type class (e.g. a_muon-CC)

    Parameters
    ----------
    class_name : str
        Particle type that should be plotted, e.g. a_muon-CC.
    mc_info : ndarray(ndim=2)
        Structured array containing the MC info.
    nn_pred_correct : ndarray(ndim=1)
        Boolean array which specifies, if the predictions of the nn for each event are correct or not.
    axes : mpl.axes
        Matplotlib axes object.
    invert : bool
        If True, it inverts the y-axis which may be useful for plotting a
        'Fraction of events classified as track' plot.
    plot_range : tuple(int,int)
        Tuple that specifies the X-Range of the plot.
    linestyle : str
         Specifies the mpl linestyle that should be used.
    color : str
        Specifies the mpl color that should be used for plotting the step plot.

    """
    ptype, is_cc = mc_info['particle_type'], mc_info['is_cc']

    is_class = select_class(class_name, ptype, is_cc)

    energy = mc_info[is_class]['energy']
    if energy.size == 0:
        return  # class not contained
    nn_pred_correct = nn_pred_correct[is_class].astype(np.int8)

    hist_1d_energy = np.histogram(energy, bins=99, range=plot_range)
    # e-hist for only correctly classified events
    hist_1d_energy_pred_correct = np.histogram(energy[nn_pred_correct == 1], bins=99, range=plot_range)

    bin_edges = hist_1d_energy[1]
    hist_1d_energy_accuracy_bins = np.divide(hist_1d_energy_pred_correct[0], hist_1d_energy[0], dtype=np.float64)  # TODO solve division by zero

    # This function makes plots with a "Fraction of events classified as track" Y-axis
    # , so need to invert the hist_1d_energy_accuracy for shower events!
    if invert is True:
        hist_1d_energy_accuracy_bins = np.absolute(hist_1d_energy_accuracy_bins - 1)

    # For making it work with matplotlib step plot
    hist_1d_energy_accuracy_bins_leading_zero = np.hstack((0, hist_1d_energy_accuracy_bins))

    label = get_latex_code_for_ptype_str(class_name)

    axes.step(bin_edges, hist_1d_energy_accuracy_bins_leading_zero, where='pre', linestyle=linestyle, color=color,
              label=label, zorder=3)

# -- Functions for making energy to accuracy plots -- #


# -- Functions for making probability plots -- #

def make_ts_prob_hists(pred_file, savefolder, cuts=None):
    """
    Makes track-shower histograms for the nn class predictions (track, shower).

    In total, 2 plots will be generated:
    1) A plot which shows the nn prediction probabilities for the track class for each particle type class
    1) A plot which shows the nn prediction probabilities for the shower class for each particle type class

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the track-shower classification predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    """
    def configure_and_save_plot(plot_title, savepath):
        """
        Configure a mpl plot with GridLines, Logscale etc.

        Parameters
        ----------
        plot_title : str
            Title that should be used for the plot.
        savepath : str
            Filepath that should be used for saving the plot.

        """
        axes.legend(loc='upper center', ncol=2)
        plt.grid(True, zorder=0, linestyle='dotted')
        # plt.yscale('log')

        x_ticks_major = np.arange(0, 1.1, 0.1)
        plt.xticks(x_ticks_major)
        plt.minorticks_on()

        plt.xlabel('OrcaNet probability')
        plt.ylabel('Normed Quantity')
        title = plt.title(plot_title)
        title.set_position([.5, 1.04])

        plt.savefig(savepath + '.pdf')
        plt.savefig(savepath + '.png', dpi=600)

    fig, axes = plt.subplots()

    # make energy cut, 1-40GeV
    y_pred, mc_info = pred_file['pred'], pred_file['mc_info']
    e_lower_40 = mc_info['energy'] <= 40
    y_pred, mc_info = y_pred[e_lower_40], mc_info[e_lower_40]

    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        y_pred = y_pred[evt_sel_mask]

    for ts_class in ['track', 'shower']:
        make_prob_hist_class('muon-CC', mc_info, y_pred, ts_class, axes, color='b', linestyle='-')
        make_prob_hist_class('a_muon-CC', mc_info, y_pred, ts_class, axes, color='b', linestyle='--')
        make_prob_hist_class('elec-CC', mc_info, y_pred, ts_class, axes, color='r', linestyle='-')
        make_prob_hist_class('a_elec-CC', mc_info, y_pred, ts_class, axes, color='r', linestyle='--')
        make_prob_hist_class('elec-NC', mc_info, y_pred, ts_class, axes, color='saddlebrown', linestyle='-')
        make_prob_hist_class('a_elec-NC', mc_info, y_pred, ts_class, axes, color='saddlebrown', linestyle='--')
        make_prob_hist_class('tau-CC', mc_info, y_pred, ts_class, axes, color='g', linestyle='-')
        make_prob_hist_class('a_tau-CC', mc_info, y_pred, ts_class, axes, color='g', linestyle='--')
        configure_and_save_plot(plot_title='Probability to be classified as ' + ts_class + ', 3-40GeV',
                                savepath=savefolder + '/ts_prob_' + ts_class)
        plt.cla()

    plt.close()


def make_prob_hist_class(class_name, mc_info, y_pred, ts_class, axes, plot_range=(0, 1), color='b', linestyle='-'):
    """
    Adds a mpl ts probability hist to an axes object for a certain particle class (e.g. 'muon-CC').

    The X-axis of the hist is the probability for the ts_class, and the Y-axis shows the normed quantity.

    Parameters
    ----------
    class_name : str
        Particle type that should be plotted in the hist, e.g. a_muon-CC.
    mc_info : ndarray(ndim=2)
        Structured array containing the MC info.
    y_pred : ndarray(ndim=2)
        Structured array containing the predicted nn probabilities for the shower and the track class.
    ts_class : str
        Specifies for which ts class ("shower", "track") the nn prediction probability should be plotted.
    axes : mpl.axes
        Matplotlib axes object.
    plot_range : tuple(int,int)
        Tuple that specifies the X-Range of the plot.
    color : str
        Specifies the mpl color that should be used for plotting the hist.
    linestyle : str
        Specifies the mpl linestyle that should be used.

    """
    ptype, is_cc = mc_info['particle_type'], mc_info['is_cc']
    is_class = select_class(class_name, ptype, is_cc)

    nn_prob_class = y_pred[is_class]['prob_' + ts_class]
    if nn_prob_class.size == 0:
        return  # class not contained

    label = get_latex_code_for_ptype_str(class_name)

    axes.hist(nn_prob_class, bins=40, range=plot_range, density=True, color=color, label=label,
              histtype='step', linestyle=linestyle, zorder=3)

# -- Functions for making probability plots -- #


def plot_ts_separability(pred_file, savefolder, pred_file_2=None, cuts=None):
    """
    Calculates and plots the separability (1-correlation coefficient) plot for ts classifiers.

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the track-shower classification predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    pred_file_2 : h5py.File
        H5py file instance of the predictions of a second classifier, that should be plotted
        into the same plot, e.g. a shallow learning classifier as comparison.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    """
    mc_info = pred_file['mc_info']
    prob_track = pred_file['prob_track']

    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        prob_track = prob_track[evt_sel_mask]

    separabilities = calculcate_separability(mc_info, prob_track)

    # plot the correlation coefficients
    fig, axes = plt.subplots()
    plt.plot(separabilities[:, 1], separabilities[:, 0], 'b', marker='o', lw=0.5, markersize=3,
             label='Deep Learning')

    if pred_file_2 is not None:
        mc_info_2 = pred_file_2['mc_info']
        prob_track_2 = pred_file_2['prob_track']

        if cuts is not None:
            assert isinstance(cuts, str)
            evt_sel_mask_2 = get_event_selection_mask(mc_info_2, cut_name=cuts)
            mc_info_2 = mc_info_2[evt_sel_mask_2]
            prob_track_2 = prob_track[evt_sel_mask_2]

        separabilities_2 = calculcate_separability(mc_info_2, prob_track_2)
        plt.plot(separabilities_2[:, 1], separabilities_2[:, 0], 'r', marker='o', lw=0.5,
                 markersize=3, label='Shallow Learning')

    plt.xlabel('Energy [GeV]')
    plt.ylabel('Separability (1-c)')
    plt.grid(True, zorder=0, linestyle='dotted')

    axes.legend(loc='center right')
    title = plt.title('Separability for track vs shower PID')
    title.set_position([.5, 1.04])

    plt.yticks(np.arange(0, 1.1, 0.1))
    # plt.xticks(np.arange(0, 110, 10))
    plt.xscale('log')
    plt.text(0.05, 0.92, 'KM3NeT Preliminary', transform=axes.transAxes, weight='bold')

    plt.savefig(savefolder + '/ts_correlation_coefficients.pdf')
    plt.savefig(savefolder + '/ts_correlation_coefficients.png', dpi=600)

    plt.close()


def calculcate_separability(mc_info, prob_track, bins=40, e_cut_range=np.logspace(0.3, 2, 18)):
    """
    Calculates the separability per energy bin by calculating the correlation factors (s=1-c).

    Parameters
    ----------
    mc_info : h5py.dataset.Dataset/ndarray(ndim=2)
        The mc_info dataset of an OrcaNet nn prediction file.
    prob_track : ndarray(ndim=1)
        Array containing the track probabilities by a classifier.
    bins : int
        For how many energy bins the separability should be calculated.
    e_cut_range : ndarray(ndim=1)
        Energy range for the calculation of the separability factors.

    """
    mc_energy = mc_info['energy']
    particle_type, is_cc = mc_info['particle_type']['is_cc']
    is_muon_cc = select_ic(particle_type, is_cc, 'muon-CC')
    is_elec_cc = select_ic(particle_type, is_cc, 'elec-CC')

    n = 0
    separabilities = []
    for e_cut in zip(e_cut_range[:-1], e_cut_range[1:]):
        n += 1
        if n <= 2:
            continue  # ecut steffen

        e_cut_mask = np.logical_and(e_cut[0] <= mc_energy, mc_energy <= e_cut[1])

        is_muon_cc_and_e_cut = np.logical_and(is_muon_cc, e_cut_mask)
        is_elec_cc_and_e_cut = np.logical_and(is_elec_cc, e_cut_mask)

        hist_prob_track_e_cut_muon_cc = np.histogram(prob_track[is_muon_cc_and_e_cut], bins=40, density=True)
        hist_prob_track_e_cut_elec_cc = np.histogram(prob_track[is_elec_cc_and_e_cut], bins=40, density=True)

        correlation_coeff_enumerator = 0
        for j in range(bins - 1):
            correlation_coeff_enumerator += hist_prob_track_e_cut_muon_cc[0][j] * hist_prob_track_e_cut_elec_cc[0][j]

        sum_prob_muon_cc = np.sum(hist_prob_track_e_cut_muon_cc[0] ** 2)
        sum_prob_elec_cc = np.sum(hist_prob_track_e_cut_elec_cc[0] ** 2)

        correlation_coeff_denominator = np.sqrt(sum_prob_muon_cc * sum_prob_elec_cc)

        separability = 1 - correlation_coeff_enumerator / float(correlation_coeff_denominator)

        average_energy = 10 ** ((np.log10(e_cut[1]) + np.log10(e_cut[0])) / float(2))

        separabilities.append((separability, average_energy))

    separabilities = np.array(separabilities)

    return separabilities



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Scripts for making specific models.
"""

import keras as ks
from keras.layers import Concatenate, Flatten, BatchNormalization, Dropout
from keras.models import Model
import toml

from orcanet.builder_util.builders import BlockBuilder


class ModelBuilder:
    """
    Can build models adapted to an Organizer instance from a toml file.

    The input of the model will match the dimensions of the input
    data given to the Organizer taking into account the sample
    modifier.

    Attributes
    ----------
    body_arch : str
        Name of the architecture of the model (see self.build).
    body_configs : list
        List with keywords for building each layer block in the model.
    body_args : dict
        Default values for the layer blocks in the model.
    head_arch : str
        Determines the head architecture of the model.
    head_arch_args : dict
        Keyword arguments for the specific output_arch used.
    head_args : dict
        Default values for the head layer blocks in the model.
    optimizer : str
        Optimizer for training the model. Either "Adam" or "SGD".
    compile_opt : dict
        Keys: Names of the output layers of the model.
        Values: Loss function, optionally weight and optionally metric of
        each output layer.
        Format: { layer_name : { loss_function:, weight:, metrics: } }
        The loss_function is a string or a function, the weight is a float
        and metrics is a list of functions/strings.

    """
    def __init__(self, model_file):
        """
        Read out parameters for creating models with OrcaNet from a toml file.

        Parameters
        ----------
        model_file : str
            Path to the model toml file.

        """
        file_content = toml.load(model_file)

        try:
            body = file_content["body"]
            self.body_arch = body.pop("architecture")
            self.body_configs = body.pop('blocks')
            self.body_args = body

            head = file_content["head"]
            self.head_arch = head.pop("architecture")
            self.head_arch_args = head.pop("architecture_args")
            self.head_args = head

            compile_sect = file_content["compile"]
            self.optimizer = compile_sect["optimizer"]
            self.compile_opt = compile_sect["losses"]

        except KeyError as e:
            if len(e.args) == 1:
                option = e.args[0]
            else:
                option = e.args
            raise KeyError("Missing parameter in toml model file: "
                           + str(option)) from None

    def build(self, orga):
        """
        Build the network.

        Input layers will be adapted to the input files in the organizer.
        Can also add the matching modifiers and custom objects to the orga.

        Parameters
        ----------
        orga : object Organizer
            Contains all the configurable options in the OrcaNet scripts.

        Returns
        -------
        model : keras model
            The network.

        """
        input_shapes = orga.io.get_input_shapes()
        custom_objects = orga.cfg.custom_objects

        if self.body_arch == "single":
            # Build a single input network
            if len(input_shapes) is not 1:
                raise ValueError("Invalid input_shape for architecture {}: "
                                 "Has length {}, but must be 1\n input_shapes "
                                 "= {}".format(self.body_arch, len(input_shapes),
                                               input_shapes))
            builder = BlockBuilder(self.body_args, self.head_args)
            model = builder.build(input_shapes, self.body_configs,
                                  self.head_arch, self.head_arch_args)

        elif self.body_arch == "multi":
            # Build an identical network up to flatten for all inputs,
            # then concatenate and add head layers
            builder = BlockBuilder(self.body_args, self.head_args)
            mid_layers, input_layers = [], []
            for input_name, input_shape in input_shapes.items():
                model = builder.build({input_name: input_shape},
                                      self.body_configs, head_arch=None)
                assert len(model.inputs) == 1, "model input is not length 1 " \
                                               "{}".format(model.inputs)
                assert len(model.outputs) == 1, "model output is not length 1 " \
                                                "{}".format(model.outputs)
                input_layers.append(model.inputs[0])
                mid_layers.append(model.outputs[0])
            x = Concatenate()(mid_layers)
            output_layer = builder.attach_output_layers(x, self.head_arch,
                                                        self.head_arch_args)
            model = Model(input_layers, output_layer)

        elif self.body_arch == "merge":
            # Concatenate multiple models to a single big one.
            # block_config is a list of paths
            model_list = []
            for path in self.body_configs:
                model_list.append(ks.models.load_model(path))
            model = self.merge_models(model_list)

        else:
            raise NameError("Unknown architecture: ", self.body_arch)

        """
        if n_gpu is not None:
            model, orga.cfg.batchsize = parallelize_model(model, n_gpu,
                                                          orga.cfg.batchsize)
        """
        self.compile_model(model, custom_objects)
        model.summary()
        return model

    def merge_models(self, model_list, trainable=False, stateful=True,
                     no_drop=True):
        """
        Concatenate two or more single input cnns to a big one.

        It will explicitly look for a Flatten layer and cut after it,
        Concatenate all models, and then add the head layers.

        Parameters
        ----------
        model_list : list
            List of keras models to stitch together.
        trainable : bool
            Whether the layers of the loaded models will be trainable.
        stateful : bool
            Whether the batchnorms of the loaded models will be stateful.
        no_drop : bool
            If true, rate of dropout layers from loaded models will
            be set to zero.

        Returns
        -------
        model : keras model
            The uncompiled merged keras model.

        """
        # Get the input and Flatten layers in each of the given models
        input_layers, flattens = [], []
        for model in model_list:
            if len(model.inputs) != 1:
                raise ValueError(
                    "model input is not length 1 {}".format(model.inputs))
            input_layers.append(model.input)
            flatten_found = 0
            for layer in model.layers:
                layer.trainable = trainable
                if isinstance(layer, BatchNormalization):
                    layer.stateful = stateful
                elif isinstance(layer, Flatten):
                    flattens.append(layer.output)
                    flatten_found += 1
            if flatten_found != 1:
                raise TypeError(
                    "Expected 1 Flatten layer but got " + str(flatten_found))

        # attach new head
        x = Concatenate()(flattens)
        builder = BlockBuilder(body_defaults=None,
                               head_defaults=self.head_args)
        output_layer = builder.attach_output_layers(x, self.head_arch,
                                                    self.head_arch_args)

        model = Model(input_layers, output_layer)
        if no_drop:
            model = change_dropout_rate(model, before_concat=0.)

        return model

    def compile_model(self, model, custom_objects=None):
        """
        Compile a model with the optimizer settings given as the attributes.

        Parameters
        ----------
        model : ks.model
            A keras model.
        custom_objects : dict or None
            Maps names (strings) to custom loss functions.

        Returns
        -------
        model : keras model
            The compiled (or recompiled) keras model.

        """
        if self.optimizer == 'adam':
            optimizer = ks.optimizers.Adam(beta_1=0.9, beta_2=0.999, epsilon=0.1,
                                           decay=0.0)  # epsilon=1 for deep networks
        elif self.optimizer == 'sgd':
            optimizer = ks.optimizers.SGD(momentum=0.9, decay=0, nesterov=True)
        else:
            raise NameError('Unknown optimizer name ({})'.format(self.optimizer))

        loss_functions, loss_weights, loss_metrics = {}, {}, {}
        for layer_name, layer_info in self.compile_opt.items():
            # Replace the str function name with actual function if it is custom
            loss_function = layer_info['function']
            if custom_objects is not None and loss_function in custom_objects:
                loss_function = custom_objects[loss_function]
            loss_functions[layer_name] = loss_function

            # Use given weight, else use default weight of 1
            if 'weight' in layer_info:
                weight = layer_info['weight']
            else:
                weight = 1.0
            loss_weights[layer_name] = weight

            # Use given metrics, else use no metrics
            if 'metrics' in layer_info:
                metrics = layer_info['metrics']
            else:
                metrics = []
            loss_metrics[layer_name] = metrics

        model.compile(loss=loss_functions, optimizer=optimizer,
                      metrics=loss_metrics, loss_weights=loss_weights)
        return model


def change_dropout_rate(model, before_concat, after_concat=None):
    """
    Change the dropout rate in a model.

    Only for models with a concatenate layer, aka multiple
    single input models that were merged together.

    Parameters
    ----------
    model : keras model

    before_concat : float
        New dropout rate before the concatenate layer in the model.
    after_concat : float or None
        New dropout rate after the concatenate layer. None will leave the
        dropout rate there as it was.

    """
    ch_bef, ch_aft, concat_found = 0, 0, 0

    for layer in model.layers:
        if isinstance(layer, Dropout):
            if concat_found == 0:
                layer.rate = before_concat
                ch_bef += 1
            else:
                layer.rate = after_concat
                ch_aft += 1

        elif isinstance(layer, Concatenate):
            concat_found += 1
            if after_concat is None:
                break

    if concat_found != 1:
        raise TypeError("Expected 1 Flatten layer but got " + str(concat_found))
    clone = ks.models.clone_model(model)
    clone.set_weights(model.get_weights())
    print("Changed dropout rates of {} layers before and {} layers after "
          "Concatenate.".format(ch_bef, ch_aft))
    return clone


# def parallelize_model(model, n_gpu, batchsize, mode="avolkov"):
#     """
#     Parallelizes the nn-model to multiple gpu's.
#
#     Currently, up to 4 GPU's at Tiny-GPU are supported.
#
#     Parameters
#     ----------
#     model : ks.model.Model
#         Keras model of a neural network.
#     n_gpu : int
#         Number of gpu's that the model should be parallelized to.
#     batchsize : int
#         Batchsize that is used for the training / inferencing of the cnn.
#     mode : str
#         Avolkov or keras.
#
#     Returns
#     -------
#     model : ks.models.Model
#         The parallelized Keras nn instance (multi_gpu_model).
#     batchsize : int
#         The new batchsize scaled by the number of used gpu's.
#
#     """
#     assert n_gpu > 1 and isinstance(n_gpu, int), 'n_gpu must be an int with n_gpu >= 1!'
#
#     if mode == "avolkov":
#         from orcanet.builder_util.multi_gpu.multi_gpu import (
#             get_available_gpus, make_parallel, print_mgpu_modelsummary)
#
#         gpus_list = get_available_gpus(n_gpu)
#         ngpus = len(gpus_list)
#         print('Using GPUs: {}'.format(', '.join(gpus_list)))
#
#         # Data-Parallelize the model via function
#         model = make_parallel(model, gpus_list, usenccl=False, initsync=True,
#                               syncopt=False, enqueue=False)
#         print_mgpu_modelsummary(model)
#         batchsize = batchsize * ngpus
#
#     elif mode == "keras":
#         # For keras, one has to save the original model, not the saved one...
#         model = ks.utils.multi_gpu_model(model, n_gpu)
#         batchsize *= n_gpu
#
#     else:
#         raise NameError("Unknown mode", mode)
#
#     return model, batchsize



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utility functions used for training a NN."""

import numpy as np
import h5py
import os
import keras as ks
from functools import reduce
import keras.backend as K


def get_auto_label_modifier(model):
    """
    Get a label_modifier for when none is specified by the user.

    Will simply assume that for every output of the model,
    there is a column in the mc_info with the same name.

    Parameters
    ----------
    model : ks.Model
        A keras model.

    Returns
    -------
    label_modifier : function

    """
    names = model.output_names

    def label_modifier(y_values):
        ys = {name: y_values[name] for name in names}
        return ys
    return label_modifier


def get_inputs(model):
    """
    Get the names and the layers of the inputs of the model.

    Parameters
    ----------
    model : ks.model
        A keras model.

    Returns
    -------
    layers :dict
        The input layers and names.

    """
    from keras.layers import InputLayer
    layers = {}
    for layer in model.layers:
        if isinstance(layer, InputLayer):
            layers[layer.name] = layer
    return layers


def get_layer_output(model, samples, layer_name, mode="test"):
    """
    Get the output of an intermediate layer from a model.

    Parameters
    ----------
    model : keras model
        The model.
    samples : List or ndarray
        The data to apply the model to.
    layer_name : str
        Name of the layer.
    mode : str
        Mode of the layers during the forward pass. Either train or test.
        Important for batchnorm, dropout, ...

    Returns
    -------
    layer_output : ndarray
        The output from the layer.

    """
    if mode == "test":
        phase = 0
    elif mode == "train":
        phase = 1
    else:
        raise NameError("Unknown mode: ", mode)

    # either a tensor or a list of tensors
    inp_tensors = model.input
    # Make input to always be a list
    if not isinstance(inp_tensors, list):
        inp_tensors = [inp_tensors, ]

    if isinstance(samples, dict):
        # doesnt work with dicts so transform into lists instead
        xs = [samples[key] for key in model.input_names]
    elif not isinstance(samples, list):
        # Make input to always be a list
        xs = [samples, ]
    else:
        xs = samples

    get_output = K.function(
        inp_tensors + [K.learning_phase(), ],
        [model.get_layer(layer_name).output])
    layer_output = get_output(xs + [phase, ])[0]
    return layer_output


# ------------- Zero center functions -------------#


def load_zero_center_data(orga, logging=False):
    """
    Gets the xs_mean array(s) that can be used for zero-centering.

    The arrays are either loaded from a previously saved .npz file or they
    are calculated on the fly by calculating the mean value per bin for the
    given training files. The name of the saved image is derived from the
    name of the list file which was given to the cfg.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    logging : bool
        If true, will log the execution of this function into the
        full summary in the output folder.

    Returns
    -------
    xs_mean : dict
        Dict of ndarray(s) that contains the mean_image of the x dataset
        (1 array per list input). Can be used for zero-centering later on.
        Example format:
        { "input_A" : ndarray, "input_B" : ndarray }

    """
    all_train_files = orga.io.get_local_files("train")
    zero_center_folder = orga.cfg.zero_center_folder
    if not zero_center_folder.endswith("/"):
        zero_center_folder += "/"
    train_files_list_name = os.path.basename(orga.cfg.get_list_file())
    key_samples = orga.cfg.key_samples

    orga.io.print_log("Zero centering", logging)
    orga.io.print_log("--------------", logging)
    orga.io.print_log("Zero center folder:   " + zero_center_folder, logging)

    xs_mean = {}
    for input_key, train_filepaths in all_train_files.items():
        xs_mean_path = get_xs_mean_path(zero_center_folder, train_filepaths)

        if xs_mean_path is not None:
            orga.io.print_log('{}:   Loading saved zero centering'.format(
                input_key), logging)
            xs_mean_ip_i = np.load(xs_mean_path)["xs_mean"]
            orga.io.print_log("\tLoaded file: {}".format(
                os.path.basename(xs_mean_path)), logging)

        else:
            orga.io.print_log('{}:   Making new zero centering'.format(
                input_key), logging)

            xs_mean_ip_i = make_xs_mean(train_filepaths, key_samples)
            filename = zero_center_folder + train_files_list_name \
                                          + '_input_' + str(input_key) + '.npz'
            np.savez(filename, xs_mean=xs_mean_ip_i,
                     zero_center_used_ip_files=train_filepaths)

            orga.io.print_log('\tSaved as {} with shape {}'.format(
                os.path.basename(filename), xs_mean_ip_i.shape), logging)

        xs_mean[input_key] = xs_mean_ip_i

    orga.io.print_log("", logging)
    return xs_mean


def get_xs_mean_path(zero_center_folder, train_filepaths):
    """
    Search for precalculated xs_mean arrays in the zero_center_folder.

    The function opens every .npz file in the zero center folder and checks
    if the files used to generate this xs_mean (stored as subarray
    'zero_center_used_ip_files') is the same as the given train_filepaths.

    Parameters
    ----------
    zero_center_folder : str
        Full path to the folder where the zero_centering arrays are stored.
    train_filepaths : list
        The filepaths of all train_files.

    Returns
    -------
    xs_mean_path : None/ndarray
        The zero center filepath for the given train_filepaths if
        it exists in the zero_center_files. If not, returns None.

    """
    xs_mean_path = None

    if not os.path.isdir(zero_center_folder):
        os.mkdir(zero_center_folder)

    for file in os.listdir(zero_center_folder):
        if not file.endswith('.npz'):
            continue
        file = zero_center_folder + file
        used_ip_files = np.load(file)['zero_center_used_ip_files']
        if np.array_equal(used_ip_files, train_filepaths):
            xs_mean_path = file
            break

    return xs_mean_path


def make_xs_mean(filepaths, key_samples, total_memory=4e9):
    """
    Calculates the zero center image of a dataset.

    Calculating still works if xs is larger than the available memory
    and also if the file is compressed.

    Parameters
    ----------
    filepaths : List
        Filepaths of the data files with the samples for which the
        mean_image will be calculated.
    key_samples : str
        The name of the datagroup in your h5 input files which contains
        the samples to the network.
    total_memory : int
        check available memory and divide the mean calculation in steps
        total_memory = 4e9  # * n_gpu # In bytes.
        Take max. 1/2 of what is available per GPU (16G), just to make sure.

    Returns
    -------
    xs_mean : ndarray
        The zero center image.

    """
    xs_means = []
    file_sizes = []

    for filepath in filepaths:

        with h5py.File(filepath, "r") as file:
            filesize = get_array_memsize(file['x'])
            steps = int(np.ceil(filesize/total_memory))
            n_rows = file[key_samples].shape[0]
            stepsize = int(n_rows / float(steps))

            # create xs_mean_arr that stores intermediate mean_temp results
            xs_mean_arr = np.zeros((steps, ) + file['x'].shape[1:],
                                   dtype=np.float64)
            print("\tCalculating for file: " + filepath)
            for i in range(steps):
                if i % 5 == 0:
                    print('\t   Step ' + str(i) + " of " + str(steps))

                # for the last step, calculate mean till the end of the file
                if i == steps-1 or steps == 1:
                    xs_mean_temp = np.mean(
                        file[key_samples][i * stepsize: n_rows],
                        axis=0, dtype=np.float64)
                else:
                    xs_mean_temp = np.mean(
                        file[key_samples][i*stepsize: (i+1) * stepsize],
                        axis=0, dtype=np.float64)

                xs_mean_arr[i] = xs_mean_temp

        print("\tDone!")
        # The mean for this file
        xs_means.append(np.mean(xs_mean_arr, axis=0,
                                dtype=np.float64).astype(np.float32))
        # the number of samples in this file
        file_sizes.append(n_rows)

    # calculate weighted average depending on no of samples in the files
    file_sizes = [size / np.sum(file_sizes) for size in file_sizes]
    xs_mean = np.average(xs_means, weights=file_sizes, axis=0)
    return xs_mean


def get_array_memsize(array):
    """
    Calculates the approximate memory size of an array.
    :param ndarray array: an array.
    :return: float memsize: size of the array in bytes.
    """
    shape = array.shape
    n_numbers = reduce(lambda x, y: x*y, shape)  # number of entries in an array
    precision = 8  # Precision of each entry, typically uint8 for xs datasets
    memsize = (n_numbers * precision) / float(8)  # in bytes

    return memsize



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import os
from keras.models import Model
from keras.layers import Dense, Input, Flatten, Convolution3D
from unittest import TestCase
from unittest.mock import MagicMock

from orcanet.core import Organizer
from orcanet.model_builder import ModelBuilder
from orcanet_contrib.custom_objects import get_custom_objects


class TestModel(TestCase):
    def setUp(self):
        """
        Make a .temp directory in the current working directory, generate
        dummy data in it and set up the cfg object.

        """
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.model_file = os.path.join(self.data_folder, "model_builder_test_CNN_cat.toml")

        self.input_shapes = {
            "input_A": (3, 3, 3, 3),
        }

        orga = Organizer(".")
        orga.io.get_input_shapes = MagicMock(return_value=self.input_shapes)
        orga.cfg.custom_objects = get_custom_objects()

        self.orga = orga

    def test_model_setup_CNN_model(self):
        orga = self.orga

        builder = ModelBuilder(self.model_file)
        model = builder.build(orga)

        self.assertEqual(model.input_shape[1:], self.input_shapes["input_A"])
        self.assertEqual(model.output_shape[1:], (2, ))
        self.assertEqual(len(model.layers), 14)

    def test_merge_models(self):
        def build_model(inp_layer_name, inp_shape):
            inp = Input(inp_shape, name=inp_layer_name)
            x = Convolution3D(3, 3)(inp)
            x = Flatten()(x)
            out = Dense(1, name="out_0")(x)

            model = Model(inp, out)
            return model

        model_file = os.path.join(self.data_folder, self.model_file)
        builder = ModelBuilder(model_file)
        model1 = build_model("inp_A", self.input_shapes["input_A"])
        model2 = build_model("inp_B", self.input_shapes["input_A"])
        merged_model = builder.merge_models([model1, model2])

        for layer in model1.layers + model2.layers:
            if isinstance(layer, Dense):
                continue
            merged_layer = merged_model.get_layer(layer.name)
            for i in range(len(layer.get_weights())):
                self.assertTrue(np.array_equal(layer.get_weights()[i],
                                               merged_layer.get_weights()[i]))
